import os
import unittest
import warnings

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import Site, _alt_az_pa_from_ra_dec, calc_lmst

import rubin_sim.maf.stackers as stackers
from rubin_sim.maf import get_sim_data


class TestStackerClasses(unittest.TestCase):
    def setUp(self):
        # get some of the test data
        test_db = os.path.join(get_data_dir(), "tests", "example_dbv1.7_0yrs.db")
        query = "select * from observations limit 1000"
        self.test_data = get_sim_data(test_db, None, [], full_sql_query=query)

    def test_stackers_run(self):
        """Just run all of the stackers with our example data."""
        for stacker_class in stackers.BaseStacker.registry.values():
            stacker = stacker_class()
            stacker_name = stacker.__class__.__name__.lower()
            if stacker_name.startswith("sdss"):
                continue
            if isinstance(stacker, stackers.BaseMoStacker):
                continue
            try:
                stacker.run(self.test_data)
            except NotImplementedError:
                pass
            except:
                print(f"Failed at stacker {stacker.__class__.__name__}")
                raise

    def test_add_cols(self):
        """Test that we can add columns as expected."""
        data = np.zeros(90, dtype=list(zip(["alt"], [float])))
        data["alt"] = np.arange(0, 90)
        stacker = stackers.ZenithDistStacker(alt_col="alt", degrees=True)
        newcol = stacker.cols_added[0]
        # First - are the columns added if they are not there.
        data, cols_present = stacker._add_stacker_cols(data)
        self.assertEqual(cols_present, False)
        self.assertIn(newcol, data.dtype.names)
        # Next - if they are present, is that information passed back?
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            data, cols_present = stacker._add_stacker_cols(data)
            self.assertEqual(cols_present, True)

    def test_eq(self):
        """
        Test that stackers can be compared
        """
        s1 = stackers.ParallaxFactorStacker()
        s2 = stackers.ParallaxFactorStacker()
        assert s1 == s2

        # Test if they have numpy array atributes
        s1.ack = np.arange(10)
        s2.ack = np.arange(10)
        assert s1 == s2

        # Change the array and test
        s1.ack += 1
        assert s1 != s2

        s2 = stackers.NormAirmassStacker()
        assert s1 != s2

    def test_norm_airmass(self):
        """
        Test the normalized airmass stacker.
        """
        rng = np.random.RandomState(232)
        data = np.zeros(600, dtype=list(zip(["airmass", "fieldDec"], [float, float])))
        data["airmass"] = rng.random_sample(600)
        data["fieldDec"] = rng.random_sample(600) * np.pi - np.pi / 2.0
        data["fieldDec"] = np.degrees(data["fieldDec"])
        stacker = stackers.NormAirmassStacker(degrees=True)
        data = stacker.run(data)
        for i in np.arange(data.size):
            self.assertLessEqual(data["normairmass"][i], data["airmass"][i])
        self.assertLess(np.min(data["normairmass"] - data["airmass"]), 0)

    def test_parallax_factor(self):
        """
        Test the parallax factor.
        """

        data = np.zeros(
            600,
            dtype=list(
                zip(
                    ["fieldRA", "fieldDec", "observationStartMJD"],
                    [float, float, float],
                )
            ),
        )
        data["fieldRA"] = data["fieldRA"] + 0.1
        data["fieldDec"] = data["fieldDec"] - 0.1
        data["observationStartMJD"] = np.arange(data.size) + 49000.0
        stacker = stackers.ParallaxFactorStacker(degrees=True)
        data = stacker.run(data)
        self.assertLess(max(np.abs(data["ra_pi_amp"])), 1.1)
        self.assertLess(max(np.abs(data["dec_pi_amp"])), 1.1)
        self.assertLess(np.max(data["ra_pi_amp"] ** 2 + data["dec_pi_amp"] ** 2), 1.1)
        self.assertGreater(min(np.abs(data["ra_pi_amp"])), 0.0)
        self.assertGreater(min(np.abs(data["dec_pi_amp"])), 0.0)

    def _t_dither_range(self, diffsra, diffsdec, ra, dec, max_dither):
        self.assertLessEqual(np.abs(diffsra).max(), max_dither)
        self.assertLessEqual(np.abs(diffsdec).max(), max_dither)
        offsets = np.sqrt(diffsra**2 + diffsdec**2)
        self.assertLessEqual(offsets.max(), max_dither)
        self.assertGreater(diffsra.max(), 0)
        self.assertGreater(diffsdec.max(), 0)
        self.assertLess(diffsra.min(), 0)
        self.assertLess(diffsdec.min(), 0)

    def _t_dither_per_night(self, diffsra, diffsdec, ra, dec, nights):
        n = np.unique(nights)
        for ni in n:
            match = np.where(nights == ni)[0]
            dra_on_night = np.abs(np.diff(diffsra[match]))
            ddec_on_night = np.abs(np.diff(diffsdec[match]))
            if dra_on_night.max() > 0.0005:
                print(ni)
                m = np.where(dra_on_night > 0.0005)[0]
                print(diffsra[match][m])
                print(ra[match][m])
                print(dec[match][m])
                print(dra_on_night[m])
            self.assertAlmostEqual(dra_on_night.max(), 0)
            self.assertAlmostEqual(ddec_on_night.max(), 0)

    @unittest.skip("Dither Stackers deprecated")
    def test_setup_dither_stackers(self):
        # Test that we get no stacker when using default columns.
        ra_col = "fieldRA"
        dec_col = "fieldDec"
        degrees = True
        stackerlist = stackers.setup_dither_stackers(ra_col, dec_col, degrees)
        self.assertEqual(len(stackerlist), 0)
        # Test that we get one (and the right one)
        # when using particular columns.
        ra_col = "hexDitherFieldPerNightRa"
        dec_col = "hexDitherFieldPerNightDec"
        stackerlist = stackers.setup_dither_stackers(ra_col, dec_col, degrees)
        self.assertEqual(len(stackerlist), 1)
        self.assertEqual(stackerlist[0], stackers.HexDitherFieldPerNightStacker())
        # Test that kwargs are passed along.
        stackerlist = stackers.setup_dither_stackers(ra_col, dec_col, degrees, max_dither=0.5)
        self.assertEqual(stackerlist[0].max_dither, np.radians(0.5))

    @unittest.skip("Dither Stackers deprecated")
    def test_base_dither_stacker(self):
        # Test that the base dither stacker matches the type of a stacker.
        s = stackers.HexDitherFieldPerNightStacker()
        self.assertTrue(isinstance(s, stackers.BaseDitherStacker))
        s = stackers.ParallaxFactorStacker()
        self.assertFalse(isinstance(s, stackers.BaseDitherStacker))

    @unittest.skip("Dither Stackers deprecated")
    def test_random_dither(self):
        """
        Test the random dither pattern.
        """
        max_dither = 0.5
        data = np.zeros(600, dtype=list(zip(["fieldRA", "fieldDec"], [float, float])))
        # Set seed so the test is stable
        rng = np.random.RandomState(42)
        # Restrict dithers to area where wraparound is not a problem for
        # comparisons.
        data["fieldRA"] = np.degrees(rng.random_sample(600) * (np.pi) + np.pi / 2.0)
        data["fieldDec"] = np.degrees(rng.random_sample(600) * np.pi / 2.0 - np.pi / 4.0)
        stacker = stackers.RandomDitherFieldPerVisitStacker(max_dither=max_dither)
        data = stacker.run(data)
        diffsra = (data["fieldRA"] - data["randomDitherFieldPerVisitRa"]) * np.cos(
            np.radians(data["fieldDec"])
        )
        diffsdec = data["fieldDec"] - data["randomDitherFieldPerVisitDec"]
        # Check dithers within expected range.
        self._t_dither_range(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], max_dither)

    @unittest.skip("Dither Stackers deprecated")
    def test_random_dither_per_night(self):
        """
        Test the per-night random dither pattern.
        """
        max_dither = 0.5
        ndata = 600
        # Set seed so the test is stable
        rng = np.random.RandomState(42)

        data = np.zeros(
            ndata,
            dtype=list(
                zip(
                    ["fieldRA", "fieldDec", "fieldId", "night"],
                    [float, float, int, int],
                )
            ),
        )
        data["fieldRA"] = rng.rand(ndata) * (np.pi) + np.pi / 2.0
        data["fieldDec"] = rng.rand(ndata) * np.pi / 2.0 - np.pi / 4.0
        data["fieldId"] = np.floor(rng.rand(ndata) * ndata)
        data["night"] = np.floor(rng.rand(ndata) * 10).astype("int")
        stacker = stackers.RandomDitherPerNightStacker(max_dither=max_dither)
        data = stacker.run(data)
        diffsra = (np.radians(data["fieldRA"]) - np.radians(data["randomDitherPerNightRa"])) * np.cos(
            np.radians(data["fieldDec"])
        )
        diffsdec = np.radians(data["fieldDec"]) - np.radians(data["randomDitherPerNightDec"])
        self._t_dither_range(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], max_dither)
        # Check that dithers on the same night are the same.
        self._t_dither_per_night(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], data["night"])

    @unittest.skip("Dither Stackers deprecated")
    def test_spiral_dither_per_night(self):
        """
        Test the per-night spiral dither pattern.
        """
        max_dither = 0.5
        ndata = 2000
        # Set seed so the test is stable
        rng = np.random.RandomState(42)

        data = np.zeros(
            ndata,
            dtype=list(
                zip(
                    ["fieldRA", "fieldDec", "fieldId", "night"],
                    [float, float, int, int],
                )
            ),
        )
        data["fieldRA"] = rng.rand(ndata) * (np.pi) + np.pi / 2.0
        data["fieldRA"] = np.zeros(ndata) + np.pi / 2.0
        data["fieldDec"] = rng.rand(ndata) * np.pi / 2.0 - np.pi / 4.0
        data["fieldDec"] = np.zeros(ndata)
        data["fieldId"] = np.floor(rng.rand(ndata) * ndata)
        data["night"] = np.floor(rng.rand(ndata) * 20).astype("int")
        stacker = stackers.SpiralDitherPerNightStacker(max_dither=max_dither)
        data = stacker.run(data)
        diffsra = (data["fieldRA"] - data["spiralDitherPerNightRa"]) * np.cos(np.radians(data["fieldDec"]))
        diffsdec = data["fieldDec"] - data["spiralDitherPerNightDec"]
        self._t_dither_range(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], max_dither)
        # Check that dithers on the same night are the same.
        self._t_dither_per_night(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], data["night"])

    @unittest.skip("Dither Stackers deprecated")
    def test_hex_dither_per_night(self):
        """
        Test the per-night hex dither pattern.
        """
        max_dither = 0.5
        ndata = 2000
        # Set seed so the test is stable
        rng = np.random.RandomState(42)

        data = np.zeros(
            ndata,
            dtype=list(
                zip(
                    ["fieldRA", "fieldDec", "fieldId", "night"],
                    [float, float, int, int],
                )
            ),
        )
        data["fieldRA"] = rng.rand(ndata) * (np.pi) + np.pi / 2.0
        data["fieldDec"] = rng.rand(ndata) * np.pi / 2.0 - np.pi / 4.0
        data["fieldId"] = np.floor(rng.rand(ndata) * ndata)
        data["night"] = np.floor(rng.rand(ndata) * 217).astype("int")
        stacker = stackers.HexDitherPerNightStacker(max_dither=max_dither)
        data = stacker.run(data)
        diffsra = (data["fieldRA"] - data["hexDitherPerNightRa"]) * np.cos(np.radians(data["fieldDec"]))
        diffsdec = data["fieldDec"] - data["hexDitherPerNightDec"]
        self._t_dither_range(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], max_dither)
        # Check that dithers on the same night are the same.
        self._t_dither_per_night(diffsra, diffsdec, data["fieldRA"], data["fieldDec"], data["night"])

    @unittest.skip("Dither Stackers deprecated")
    def test_random_rot_dither_per_filter_change_stacker(self):
        """
        Test the rotational dither stacker.
        """
        max_dither = 90
        filt = np.array(["r", "r", "r", "g", "g", "g", "r", "r"])
        rot_tel_pos = np.array([0, 0, 0, 0, 0, 0, 0, 0], float)
        # Test that have a dither in rot offset for every filter change.
        odata = np.zeros(len(filt), dtype=list(zip(["filter", "rotTelPos"], [(np.str_, 1), float])))
        odata["filter"] = filt
        odata["rotTelPos"] = rot_tel_pos
        stacker = stackers.RandomRotDitherPerFilterChangeStacker(
            max_dither=max_dither, degrees=True, random_seed=99
        )
        data = stacker.run(odata)  # run the stacker
        random_dithers = data["randomDitherPerFilterChangeRotTelPos"]
        # Check that first three visits have the same rot_tel_pos, etc.
        rot_offsets = rot_tel_pos - random_dithers
        # no dither w/o a filter change
        self.assertEqual(rot_offsets[0], 0)
        offset_changes = np.where(rot_offsets[1:] != rot_offsets[:-1])[0]
        filt_changes = np.where(filt[1:] != filt[:-1])[0]
        # dither after every filter
        np.testing.assert_array_equal(offset_changes, filt_changes)

        # now test to ensure that user-defined max_rot_angle value works
        # and that visits in between filter changes for which no offset
        # can be found are left undithered
        # (g band visits span rotator range, so can't be dithered)
        gvisits = np.where(filt == "g")
        maxrot = 30
        rot_tel_pos[gvisits[0][0]] = -maxrot
        rot_tel_pos[gvisits[0][-1]] = maxrot
        odata["rotTelPos"] = rot_tel_pos
        stacker = stackers.RandomRotDitherPerFilterChangeStacker(
            max_dither=max_dither,
            degrees=True,
            min_rot_angle=-maxrot,
            max_rot_angle=maxrot,
            random_seed=19231,
        )
        data = stacker.run(odata)
        random_dithers = data["randomDitherPerFilterChangeRotTelPos"]
        # Check that we respected the range.
        self.assertEqual(random_dithers.max(), 30)
        self.assertEqual(random_dithers.min(), -30)
        # Check that g band visits weren't dithered.
        rot_offsets = rot_tel_pos - random_dithers
        self.assertEqual(rot_offsets[gvisits].all(), 0)

    def test_ha_stacker(self):
        """Test the Hour Angle stacker"""
        data = np.zeros(100, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = np.arange(100) / 99.0 * np.pi * 2
        stacker = stackers.HourAngleStacker(degrees=True)
        data = stacker.run(data)
        # Check that data is always wrapped
        self.assertLess(np.max(data["HA"]), 12.0)
        self.assertGreater(np.min(data["HA"]), -12.0)
        # Check that HA is zero if lst == RA
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data = stacker.run(data)
        self.assertEqual(data["HA"], 0.0)
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = 20.0
        data["fieldRA"] = 20.0
        data = stacker.run(data)
        self.assertEqual(data["HA"], 0.0)
        # Check a value
        data = np.zeros(1, dtype=list(zip(["observationStartLST", "fieldRA"], [float, float])))
        data["observationStartLST"] = 0.0
        data["fieldRA"] = np.degrees(np.pi / 2.0)
        data = stacker.run(data)
        np.testing.assert_almost_equal(data["HA"], -6.0)

    def test_pa_stacker(self):
        """Test the parallacticAngleStacker"""
        data = np.zeros(
            100,
            dtype=list(
                zip(
                    [
                        "observationStartMJD",
                        "fieldDec",
                        "fieldRA",
                        "observationStartLST",
                    ],
                    [float] * 4,
                )
            ),
        )
        data["observationStartMJD"] = np.arange(100) * 0.2 + 50000
        site = Site(name="LSST")
        data["observationStartLST"] = calc_lmst(data["observationStartMJD"], site.longitude_rad)
        data["observationStartLST"] = data["observationStartLST"] * 180.0 / 12.0
        stacker = stackers.ParallacticAngleStacker(degrees=True)
        data = stacker.run(data)
        # Check values are in good range
        assert data["PA"].max() <= 180
        assert data["PA"].min() >= -180

        # Check compared to the util
        check_pa = []
        ras = np.radians(data["fieldRA"])
        decs = np.radians(data["fieldDec"])
        for ra, dec, mjd in zip(ras, decs, data["observationStartMJD"]):
            alt, az, pa = _alt_az_pa_from_ra_dec(ra, dec, mjd, site.longitude_rad, site.latitude_rad)

            check_pa.append(pa)
        check_pa = np.degrees(check_pa)

        np.testing.assert_array_almost_equal(data["PA"], check_pa, decimal=0)

    def test_galactic_stacker(self):
        """
        Test the galactic coordinate stacker
        """
        ra, dec = np.degrees(
            np.meshgrid(np.arange(0, 2.0 * np.pi, 0.1), np.arange(-np.pi / 2, np.pi / 2, 0.1))
        )
        ra = np.ravel(ra)
        dec = np.ravel(dec)
        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.GalacticStacker(ra_col="ra", dec_col="dec")
        new_data = s.run(data)
        c = SkyCoord(ra=ra * u.deg, dec=dec * u.deg).transform_to("galactic")
        expected_l, expected_b = c.l.rad, c.b.rad
        np.testing.assert_array_equal(new_data["gall"], expected_l)
        np.testing.assert_array_equal(new_data["galb"], expected_b)

        # Check that we have all the quadrants populated
        q1 = np.where((new_data["gall"] < np.pi) & (new_data["galb"] < 0.0))[0]
        q2 = np.where((new_data["gall"] < np.pi) & (new_data["galb"] > 0.0))[0]
        q3 = np.where((new_data["gall"] > np.pi) & (new_data["galb"] < 0.0))[0]
        q4 = np.where((new_data["gall"] > np.pi) & (new_data["galb"] > 0.0))[0]
        assert q1.size > 0
        assert q2.size > 0
        assert q3.size > 0
        assert q4.size > 0

    def test_ecliptic_stacker(self):
        ra, dec = np.degrees(
            np.meshgrid(np.arange(0, 2.0 * np.pi, 0.1), np.arange(-np.pi / 2, np.pi / 2, 0.1))
        )
        ra = np.ravel(ra)
        dec = np.ravel(dec)
        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.EclipticStacker(ra_col="ra", dec_col="dec", degrees=True)
        _ = s.run(data)

        data = np.zeros(ra.size, dtype=list(zip(["ra", "dec"], [float] * 2)))
        data["ra"] += ra
        data["dec"] += dec
        s = stackers.EclipticStacker(ra_col="ra", dec_col="dec", degrees=True, subtract_sun_lon=False)
        _ = s.run(data)


if __name__ == "__main__":
    unittest.main()



#################################################################################################
# Various things declared here to be imported when running artificialStructureCalculation.
# Makes the updates easier, since the constants/objects defined here are used by different
# functions, e.g. power law constants are called by GalaxyCountsMetric_extended as well as
# GalaxyCounts_withPixelCalibration. Similarly, plot_color dictionary helps maintain a consistent
# color scheme for different observing strategies.
#
# Included here:
#  * power law constants from the i-band mock catalog; based on mocks from Padilla et al.
#  * normalization constant for galaxy counts from the mock catalogs.
#  * plot_color dictionary: colors for plotting results from different observing strategies.s
#
# Humna Awan
# humna.awan@rutgers.edu
#################################################################################################
#################################################################################################
# Power law constants for each z-bin based on N. D. Padilla et al.'s mock catalogs
# General power law form: 10**(a*m+b)
# Declare the dictionary for the power law constants
from collections import OrderedDict

power_law_const_a = OrderedDict()
power_law_const_b = OrderedDict()
# 0.<z<0.15
power_law_const_a["0.<z<0.15"] = 0.166  # July 18 Catalog
power_law_const_b["0.<z<0.15"] = -0.989  # July 18 Catalog

# 0.15<z<0.37
power_law_const_a["0.15<z<0.37"] = 0.212  # July 18 Catalog
power_law_const_b["0.15<z<0.37"] = -1.540  # July 18 Catalog

# 0.37<z<0.66
power_law_const_a[" 0.37<z<0.66"] = 0.250  # July 18 Catalog
power_law_const_b[" 0.37<z<0.66"] = -2.242  # July 18 Catalog

# 0.66<z<1.0
power_law_const_a["0.66<z<1.0"] = 0.265  # July 18 Catalog
power_law_const_b["0.66<z<1.0"] = -2.660  # July 18 Catalog

# 1.0<z<1.5
power_law_const_a["1.0<z<1.5"] = 0.283  # July 18 Catalog
power_law_const_b["1.0<z<1.5"] = -3.242  # July 18 Catalog

# 1.5<z<2.0
power_law_const_a["1.5<z<2.0"] = 0.308  # July 18 Catalog
power_law_const_b["1.5<z<2.0"] = -4.169  # July 18 Catalog

# 2.0<z<2.5
power_law_const_a["2.0<z<2.5"] = 0.364  # July 18 Catalog
power_law_const_b["2.0<z<2.5"] = -6.261  # July 18 Catalog

# 2.5<z<3.0
power_law_const_a["2.5<z<3.0"] = 0.346  # July 18 Catalog
power_law_const_b["2.5<z<3.0"] = -5.661  # July 18 Catalog

# 3.0<z<3.5
power_law_const_a["3.0<z<3.5"] = 0.381  # July 18 Catalog
power_law_const_b["3.0<z<3.5"] = -6.810  # July 18 Catalog

# 3.5<z<4.0
power_law_const_a["3.5<z<4.0"] = 0.424  # July 18 Catalog
power_law_const_b["3.5<z<4.0"] = -8.362  # July 18 Catalog

# 4.0<z<15.0
power_law_const_a["4.0<z<15.0"] = 0.462  # July 18 Catalog
power_law_const_b["4.0<z<15.0"] = -9.474  # July 18 Catalog

#################################################################################################
#################################################################################################
# Normalization to match the total galaxies from  mock catalogs to match up to CFHTLS counts for
# i<25.5 galaxy catalog. Using July 18-19 i-band mock catalogs.
normalization_constant = 3.70330701802

#################################################################################################
#################################################################################################
# dictionary to hold colors for overplotted plots.
# coaddAnalysis and OSBiasAnalysis need an ordered dictionary; aritficialStructure just needs
# colors, in no specific order.
plot_color = OrderedDict()
plot_color["NoDither"] = [0.0, 0.0, 0.0]  # black

plot_color["RandomDitherFieldPerVisit"] = [
    0.0,
    0.0,
    255 / 255.0,
]  # blue [255/255., 255/255.,   0.]  # yellow
plot_color["RandomDitherFieldPerNight"] = [255 / 255.0, 0.0, 255 / 255.0]  # magenta
plot_color["RandomDitherPerNight"] = [
    255.0 / 255.0,
    0.0,
    0.0,
]  # [139/255., 0.,0.] # dark red

plot_color["RepulsiveRandomDitherFieldPerVisit"] = [
    255 / 255.0,
    105 / 255.0,
    180 / 255.0,
]  # hot pink
plot_color["RepulsiveRandomDitherFieldPerNight"] = [
    255 / 255.0,
    255 / 255.0,
    0.0,
]  # yellow
plot_color["RepulsiveRandomDitherPerNight"] = [
    147 / 255.0,
    112 / 255.0,
    219 / 255.0,
]  # medium purple

# plot_color['PentagonDiamondDitherPerSeason'] = [220/255.,  20/255.,  60/255.] # crimson
plot_color["PentagonDitherPerSeason"] = [
    75 / 255.0,
    0.0,
    130 / 255.0,
]  # indigo [255/255.,0.,255/255.]  # magenta [0., 1/255.,  34/255.]
# plot_color['SpiralDitherPerSeason'] = [4/255., 13/255.,  34/255.]

plot_color["FermatSpiralDitherFieldPerVisit"] = [
    124 / 255.0,
    252 / 255.0,
    0.0,
]  # lawngreen
plot_color["FermatSpiralDitherFieldPerNight"] = [
    0,
    206 / 255.0,
    209 / 255.0,
]  # turqoise
plot_color["FermatSpiralDitherPerNight"] = [
    34 / 255.0,
    139 / 255.0,
    34 / 255.0,
]  # forestgreen

plot_color["SequentialHexDitherFieldPerVisit"] = [
    0 / 255.0,
    255 / 255.0,
    127 / 255.0,
]  # spring green      [0.,0.,255/255.] # blue
plot_color["SequentialHexDitherFieldPerNight"] = [139 / 255.0, 0.0, 0.0]  # dark red
plot_color["SequentialHexDitherPerNight"] = [
    184 / 255.0,
    134 / 255.0,
    11 / 255.0,
]  # dark goldenrod



import unittest

from rubin_scheduler.scheduler.surveys import generate_ddf_grid


class GenerateDDFTest(unittest.TestCase):
    def testGenDDF(self):
        """
        Test the DDF grid generator over in rubin_scheduler
        """
        result = generate_ddf_grid(survey_length=0.01, verbose=False)
        assert result is not None


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import rubin_sim.maf.plots as plots


class TestNeoDistancePlotter(unittest.TestCase):
    def setUp(self):
        rng = np.random.RandomState(61723009)
        names = [
            "eclipLat",
            "eclipLon",
            "MaxGeoDist",
            "NEOHelioX",
            "NEOHelioY",
            "filter",
        ]
        types = [float] * 5
        types.append("|S1")
        npts = 100
        self.metric_values = np.zeros(npts, list(zip(names, types)))
        self.metric_values["MaxGeoDist"] = rng.rand(npts) * 2.0
        self.metric_values["eclipLat"] = rng.rand(npts)
        self.metric_values["NEOHelioX"] = rng.rand(npts) * 3 - 1.5
        self.metric_values["NEOHelioY"] = rng.rand(npts) * 3 - 1.5 + 1
        self.metric_values["filter"] = "g"

    def test_plotter(self):
        """
        Just test that it can make a figure without throwing an error.
        """
        plotter = plots.NeoDistancePlotter()
        # Need to wrap in a list because it will usually go through the
        # UniSlicer, and will thus be an array inside a 1-element masked array
        fig = plotter([self.metric_values], None, {})
        self.assertNotEqual(fig, None)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np
import pandas as pd
from rubin_scheduler.data import get_data_dir

import rubin_sim.maf as maf


class TestSNmetrics(unittest.TestCase):
    def setUp(self):
        # Make sure we can read SN lightcurve and reference info
        if not os.path.isdir(os.path.join(get_data_dir(), "maf")):
            self.skipTest("Skipping SN tests because running unit tests without full rubin_scheduler.data.")
        # Read test point data
        testfile = os.path.join(get_data_dir(), "tests", "test_simData.hdf")
        if not os.path.isfile(testfile):
            raise FileExistsError("%s not found" % testfile)
        self.simdata = {}
        with pd.HDFStore(testfile) as f:
            keys = f.keys()
            for k in keys:
                newkey = k.lstrip("/")
                self.simdata[newkey] = f.get(k).to_records()

    def test_nsn(self):
        # Set up a mock slicerPoint
        nside = 64
        slice_point = {"nside": nside, "ebv": 0.0}

        # Set up the metric
        metric = maf.SNNSNMetric(
            season=[-1],
            n_aft=8,
            n_bef=3,
            add_dust=False,
            hard_dust_cut=0.25,
            zmin=0.1,
            zmax=0.5,
            z_step=0.03,
            daymax_step=3,
            snr_min=1,
            zlim_coeff=0.95,
            sigma_c=0.04,
            bands="grizy",
            gamma_name="gamma_WFD.hdf5",
            verbose=False,
        )

        # Expected keys and results
        expected = {}
        # Muddle with m5 - very shallow
        expected["one_season_shallow"] = metric.badval
        # Very few visits
        expected["sparse_pointing"] = metric.badval
        # Very dense pointing
        expected["dense_pointing"] = np.rec.fromrecords([(5.158324, 0.276509)], names=["nSN", "zlim"])
        # These are of the same point on the sky, with/without DD fields -
        # they should be the same as the metric rejects DD visits internally
        # ---XXX except now the metric does
        # not reject visits based on notes, so not sure why these still pass

        expected["one_season_noDD"] = np.rec.fromrecords([(0.870031, 0.289703)], names=["nSN", "zlim"])
        expected["one_season_wDD"] = np.rec.fromrecords([(0.870031, 0.289703)], names=["nSN", "zlim"])
        # Here we muddle with the visit exposure time (without changing m5)
        expected["one_season_singleExp_20"] = np.rec.fromrecords([(0.879159, 0.29006)], names=["nSN", "zlim"])
        # Here we muddle with the number of visits per exposure,
        # without changing m5 or visitExpTime
        expected["one_season_singleExp_30"] = np.rec.fromrecords(
            [(0.870031, 0.289703)], names=["nSN", "zlim"]
        )

        verbose = True

        for k in self.simdata:
            res = metric.run(self.simdata[k], slice_point=slice_point)
            if verbose:
                print("")
                print(f"pointing test {k} with {len(self.simdata[k])} visits")
                try:
                    print(f"expected results {expected[k]}")
                except KeyError:
                    print(f"no expected results for {k}")
                print(f"calculated results {res}")
                print("")
            # self.assertEqual(expected[k], res)

        # And run for DD observations
        metric = maf.SNNSNMetric(
            season=[-1],
            n_aft=10,
            n_bef=4,
            add_dust=False,
            hard_dust_cut=0.25,
            zmin=0.1,
            zmax=0.5,
            z_step=0.03,
            daymax_step=3,
            snr_min=1,
            zlim_coeff=0.95,
            sigma_c=0.04,
            bands="grizy",
            gamma_name="gamma_DDF.hdf5",
            coadd_night=True,
            verbose=False,
        )
        for k in ["one_season_wDD"]:
            res = metric.run(self.simdata[k], slice_point=slice_point)
            if verbose:
                print("")
                print(f"pointing test {k} for {len(self.simdata[k])} DD-only visits")

                print(f"calculated results {res}")
                print("")


if __name__ == "__main__":
    unittest.main()



"""
Module to test that Approximate Bandpasses are in sync with the official LSST
Bandpasses from SYSENG

Note: While the LSST bandpasses list throughput values corresponding to
wavelengths in the range of 300.0-1150.0 nm, the `approximate_baseline`
directory of throughputs is created by a script manually. It is thus possible
for this directory to fall out of sync with the SYSENG values in `baseline`.
This module is intended to test whether this is happening.
"""
import os
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.phot_utils import Bandpass, PhotometricParameters


class ApproximateBandPassTest(unittest.TestCase):
    """
    Tests for the approximate Bandpasses in the throughputs directory
    """

    long_message = True

    def setUp(self):
        """Setup before tests."""
        throughputs_dir = os.path.join(get_data_dir(), "throughputs")
        self.approx_band_pass_dir = os.path.join(throughputs_dir, "approximate_baseline")
        self.approx_band_pass_dict = {}
        for kk in ["u", "g", "r", "i", "z", "y"]:
            self.approx_band_pass_dict[kk] = Bandpass()
            self.approx_band_pass_dict[kk].read_throughput(
                os.path.join(self.approx_band_pass_dir, f"total_{kk}.dat")
            )
        self.ref_band_pass_dir = os.path.join(throughputs_dir, "baseline")
        self.ref_band_pass_dict = {}
        for kk in ["u", "g", "r", "i", "z", "y"]:
            self.ref_band_pass_dict[kk] = Bandpass()
            self.ref_band_pass_dict[kk].read_throughput(
                os.path.join(self.ref_band_pass_dir, f"total_{kk}.dat")
            )
        self.error_msg = "The failure of this test indicates that the"
        " approximate bandpasses in the lsst throughputs directory do not"
        "sync up with the baseline bandpasses is throughputs. "
        " This may require running the script :"
        " throughputs.approximate_baseline/approximateBandpasses.py"

    def test_band_pass_integrals(self):
        """Test that the ratio of the quantity
        int dlambda T(lambda) = band flux for a SED proportional to $lambda$
        for the approximate bandpasses to the SYSENG band passes is 1.0 to an
        absolute tolerance hard coded to be 1.0e-4

        """
        phot_params = PhotometricParameters()

        for bn in "ugrizy":
            ref_band_pass = self.ref_band_pass_dict[bn]
            approx_band_pass = self.approx_band_pass_dict[bn]

            zp_ref = ref_band_pass.calc_zp_t(phot_params)
            zp_approx = approx_band_pass.calc_zp_t(phot_params)
            self.assertAlmostEqual(zp_ref, zp_approx, delta=1.0e-4, msg=self.error_msg)

    def test_bandpasses_indiv(self):
        """
        Test that individual transmission values at wavelengths kept in the
        approximate bandpasses match the individual transmission values in
        the reference bandpasses
        """
        for bn in "ugrizy":
            approx_band_pass = self.approx_band_pass_dict[bn]
            ref_band_pass = self.ref_band_pass_dict[bn]
            refwavelen = ref_band_pass.wavelen
            approxwavelen = approx_band_pass.wavelen
            mask = np.zeros(len(refwavelen), dtype=bool)
            for i, wave in enumerate(refwavelen):
                if wave in approxwavelen:
                    mask[i] = True
            # Assume that the wavelengths are in order
            np.testing.assert_array_almost_equal(ref_band_pass.sb[mask], approx_band_pass.sb, decimal=3)

        def tearDown(self):
            pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Calculates the stellar density based off of Juric et al 2008 and Jackson et al 2002. For use with Field Star Count metric
import math
import sys

import numpy as np
from scipy.optimize import fsolve

zsun = 25.0
rsun = 8000.0
density_rsun = 0.0364
f = 0.1


def diskprofile(R, Z, L, H):
    part1 = (-R / L) - (abs(Z + zsun) / H)
    part2 = np.exp(part1)
    part3 = np.exp(rsun / L)
    tot = density_rsun * part2 * part3
    return tot


def thindisk(R, Z):
    return diskprofile(R, Z, 2150.0, 245.0)


def thickdisk(R, Z):
    return diskprofile(R, Z, 3261.0, 743.0)


def bulge(R, Z):
    factor = 2 * (diskprofile(0, 0, 2150.0, 245.0) + diskprofile(0, 0, 3261.0, 743.0))
    distance = (R**2 + Z**2) ** 0.5
    expfunc = np.exp(-distance / 800)
    return factor * expfunc


def halo(R, Z):
    q_h = 0.64
    n_h = 2.77
    f_h = 0.001
    part1 = R**2.0 + (Z / q_h) ** 2.0
    part2 = rsun / np.power(part1, 0.5)
    part3 = np.power(part2, n_h)
    tot = density_rsun * f_h * part3
    return tot


def stellardensity(R, Z, rho=0):
    part1 = thindisk(R, Z)
    part2 = thickdisk(R, Z)
    tot_density = part1 / 1.1 + f / 1.1 * part2 + halo(R, Z) + bulge(R, Z)
    return tot_density


if __name__ == "__main__":
    print(stellardensity(float(sys.argv[1]), float(sys.argv[2])))


# Juric et al 2008
# Jackson et al 2002



__all__ = ("SkySaturationMetric",)

import numpy as np

from .base_metric import BaseMetric


class SkySaturationMetric(BaseMetric):
    """Check if the sky would saturate a visit in an exposure"""

    def __init__(self, metric_name="SkySaturation", units="#", **kwargs):
        super().__init__(col=["saturation_mag"], units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point):
        # Saturation stacker returns NaN if the sky saturates
        finite = np.isfinite(data_slice["saturation_mag"])
        result = np.size(np.where(finite == False)[0])
        return result



# Metric for kilonova detectability based on GW170817 SED used in Scolnic et
# al. 2018 and Setzer et al. 2019. The chosen detection criteria are related to
# those used in the LSST DESC white paper detectability work and the two
# references above.
#
# Contact for this code:
# christian.setzer@fysik.su.se
from pathlib import Path

from .transient_ascii_sed_metric import TransientAsciiSEDMetric

__all__ = ("GW170817DetMetric",)
base_path = Path(__file__).parent


class GW170817DetMetric(TransientAsciiSEDMetric):
    """
    Wrapper metric class for GW170817-like kilonovae based on the
    TransientAsciiSEDMetric. Defaults are set to those corresponding to similar
    detection criteria used in Scolnic et al. 2018 and Setzer et al. 2019.
    However, due to the simplified nature of transient distribution for
    computing this metric, the criteria have been altered to only include
    criteria two and three. The chosen redshift is at the approximate mean
    redshift of the detected cosmological redshift distribution shown in
    Setzer et al. 2019.

    Parameters
    -----------
    ascii_file : str, optional
        The ascii file containing the inputs for the SED. The file must
        contain three columns - ['phase', 'wave', 'flux'] -
        of phase/epoch (in days), wavelength (Angstroms), and
        flux (ergs/s/Angstrom). Default, data provided with sims_maf_contrib.
    metric_name : str, optional
        Name of the metric, can be overwritten by user or child metric.
    z: float, optional
        Cosmological redshift at which to consider observations of the
        tranisent SED. Default 0.08.
    num_filters : int, optional
        Number of filters that need to be observed for an object to be
        counted as detected. Default 2. (if num_per_lightcurve is 0, then
        this will be reset to 0).
    filter_time : float, optional
        The time within which observations in at least num_filters are
        required (in days). Default 25.0 days.
    num_phases_to_run : int, optional
        Sets the number of phases that should be checked.
        One can imagine pathological cadences where many objects pass the
        detection criteria, but would not if the observations were offset
        by a phase-shift. Default 5.
    """

    def __init__(
        self,
        ascii_file=(base_path / "../data/DECAMGemini_SED.txt").resolve(),
        metric_name="GW170817DetMetric",
        z=0.08,
        num_filters=2,
        filter_time=25.0,
        num_phases_to_run=5,
        **kwargs,
    ):
        """"""
        super(GW170817DetMetric, self).__init__(
            ascii_file=ascii_file,
            metric_name=metric_name,
            z=z,
            num_filters=num_filters,
            filter_time=filter_time,
            num_phases_to_run=num_phases_to_run,
            **kwargs,
        )



__all__ = ("filtersPerNight", "filtersWholeSurvey")

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import standard_summary


def setupMetrics(colmap, wholesurvey=False):
    metric_list = []
    caption_list = []
    # Number of filter changes (per slice - either whole survey or X nights)
    if wholesurvey:
        metric_list.append(
            metrics.NChangesMetric(
                col=colmap["filter"],
                order_by=colmap["mjd"],
                metric_name="Total Filter Changes",
            )
        )
    else:
        metric_list.append(
            metrics.NChangesMetric(
                col=colmap["filter"],
                order_by=colmap["mjd"],
                metric_name="Filter Changes",
            )
        )
    caption_list.append("Total filter changes ")
    # Minimum time between filter changes
    metric_list.append(
        metrics.MinTimeBetweenStatesMetric(change_col=colmap["filter"], time_col=colmap["mjd"])
    )
    caption_list.append("Minimum time between filter changes ")
    # Number of filter changes faster than 10 minutes
    metric_list.append(
        metrics.NStateChangesFasterThanMetric(change_col=colmap["filter"], time_col=colmap["mjd"], cutoff=10)
    )
    caption_list.append("Number of filter changes faster than 10 minutes ")
    # Number of filter changes faster than 20 minutes
    metric_list.append(
        metrics.NStateChangesFasterThanMetric(change_col=colmap["filter"], time_col=colmap["mjd"], cutoff=20)
    )
    caption_list.append("Number of filter changes faster than 20 minutes ")
    # Maximum number of filter changes faster than 10 minutes within slice
    metric_list.append(
        metrics.MaxStateChangesWithinMetric(change_col=colmap["filter"], time_col=colmap["mjd"], timespan=10)
    )
    caption_list.append("Max number of filter  changes within a window of 10 minutes ")
    # Maximum number of filter changes faster than 20 minutes within slice
    metric_list.append(
        metrics.MaxStateChangesWithinMetric(change_col=colmap["filter"], time_col=colmap["mjd"], timespan=20)
    )
    caption_list.append("Max number of filter changes within a window of 20 minutes ")
    return metric_list, caption_list


def filtersPerNight(colmap=None, runName="opsim", nights=1, extraSql=None, extraInfoLabel=None):
    """Generate a set of metrics measuring the number and rate of filter
    changes over a given span of nights.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    nights : `int`, optional
        Size of night bin to use when calculating metrics.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    # Set up sql and info_label, if passed any additional information.
    sql = ""
    info_label = "Per"
    if nights == 1:
        info_label += " Night"
    else:
        info_label += " %s Nights" % nights
    metacaption = info_label.lower()
    if (extraSql is not None) and (len(extraSql) > 0):
        sql = extraSql
        if extraInfoLabel is None:
            info_label += " %s" % extraSql
            metacaption += ", with %s selection" % extraSql
    if extraInfoLabel is not None:
        info_label += " %s" % extraInfoLabel
        metacaption += ", %s only" % extraInfoLabel
    metacaption += "."

    displayDict = {"group": "Filter Changes", "subgroup": info_label}
    summaryStats = standard_summary()

    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=nights)
    metricList, captionList = setupMetrics(colmap)
    for m, caption in zip(metricList, captionList):
        displayDict["caption"] = caption + metacaption
        bundle = mb.MetricBundle(
            m,
            slicer,
            sql,
            run_name=runName,
            info_label=info_label,
            display_dict=displayDict,
            summary_metrics=summaryStats,
        )
        bundleList.append(bundle)

    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def filtersWholeSurvey(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None):
    """Generate a set of metrics measuring the number and rate of filter
    changes over the entire survey.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    # Set up sql and info_label, if passed any additional information.
    sql = ""
    info_label = "Whole Survey"
    metacaption = "over the whole survey"
    if (extraSql is not None) and (len(extraSql) > 0):
        sql = extraSql
        if extraInfoLabel is None:
            info_label += " %s" % extraSql
            metacaption += ", with %s selction" % extraSql
    if extraInfoLabel is not None:
        info_label += " %s" % extraInfoLabel
        metacaption += ", %s only" % (extraInfoLabel)
    metacaption += "."

    displayDict = {"group": "Filter Changes", "subgroup": info_label}

    slicer = slicers.UniSlicer()
    metricList, captionList = setupMetrics(colmap)
    for m, caption in zip(metricList, captionList):
        displayDict["caption"] = caption + metacaption
        bundle = mb.MetricBundle(
            m,
            slicer,
            sql,
            run_name=runName,
            info_label=info_label,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



__all__ = ("metadata_dir",)

import argparse
import glob
import os
import shutil

import matplotlib
import numpy as np

matplotlib.use("Agg")

from . import batches as batches
from .db import ResultsDb
from .metric_bundles import MetricBundle, MetricBundleGroup
from .metrics import CountExplimMetric
from .slicers import HealpixSlicer, HealpixSubsetSlicer


def metadata_dir():
    """
    Run the metadata batch on all .db files in a directory.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--db", type=str, default=None)
    parser.add_argument(
        "--nside",
        type=float,
        default=64,
        help="nside to use for the healpix slicer and subsetslicer for metrics.",
    )
    parser.add_argument(
        "--wfd_threshold",
        type=float,
        default=750,
        help="Threshold number of visits per pointing to use to define the WFD footprint."
        "Default value of 750 corresponds to the minimum median value per pointing from the SRD.",
    )
    parser.add_argument(
        "--no_clobber",
        dest="no_clobber",
        action="store_false",
        help="Do not remove existing directory outputs",
    )
    args = parser.parse_args()

    # If runNames not given, scan for sim_name databases in current
    # directory and use those
    # Note that 'runNames' can be full path to directories

    if args.db is None:
        # Just look for any .db files in this directory
        db_files = glob.glob("*.db")
        # But remove trackingDb and results_db if they're there
        try:
            db_files.remove("trackingDb_sqlite.db")
        except ValueError:
            pass
        try:
            db_files.remove("resultsDb_sqlite.db")
        except ValueError:
            pass
    elif isinstance(args.db, str):
        db_files = [args.db]
    else:
        db_files = args.db

    sim_names = [os.path.basename(name).replace(".db", "") for name in db_files]

    for filename, sim_name in zip(db_files, sim_names):
        # Connect to the database
        colmap = batches.col_map_dict()

        # Set and create if needed the output directory
        out_dir = sim_name + "_meta"
        if not args.no_clobber:
            if os.path.isdir(out_dir):
                shutil.rmtree(out_dir)

        # Find the 'wfd' footprint
        m = CountExplimMetric(col="observationStartMJD")
        allsky_slicer = HealpixSlicer(nside=args.nside)
        constraint = 'note not like "%DD%"'
        bundle = MetricBundle(m, allsky_slicer, constraint, run_name=sim_name)
        g = MetricBundleGroup({f"{sim_name} footprint": bundle}, filename, out_dir=out_dir)
        g.run_all()
        wfd_footprint = bundle.metric_values.filled(0)
        wfd_footprint = np.where(wfd_footprint > args.wfd_threshold, 1, 0)
        wfd_hpix = np.where(wfd_footprint == 1)[0]
        wfd_slicer = HealpixSubsetSlicer(nside=args.nside, hpid=wfd_hpix)

        bdict = batches.info_bundle_dicts(allsky_slicer, wfd_slicer, sim_name, colmap)

        # Set up the resultsDB
        results_db = ResultsDb(out_dir=out_dir)
        # Go and run it
        group = MetricBundleGroup(bdict, filename, out_dir=out_dir, results_db=results_db, save_early=False)
        group.run_all(clear_memory=True, plot_now=True)
        results_db.close()



__all__ = ("wrap_ra", "robust_rms", "recalc_mags")

import glob
import os

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.phot_utils import Bandpass, Sed


def wrap_ra(ra):
    """
    Wrap only RA values into 0-2pi (using mod).
    """
    ra = ra % (2.0 * np.pi)
    return ra


def robust_rms(array, missing=0.0):
    """
    Use the interquartile range to compute a robust approximation of the RMS.
    if passed an array smaller than 2 elements, return missing value
    """
    if np.size(array) < 2:
        rms = missing
    else:
        iqr = np.percentile(array, 75) - np.percentile(array, 25)
        rms = iqr / 1.349  # approximation
    return rms


def spec2mags(spectra_list, wave):
    """Convert sky spectra to magnitudes"""
    # Load LSST filters
    through_path = os.path.join(get_data_dir(), "throughputs/baseline")
    keys = ["u", "g", "r", "i", "z", "y"]

    dtype = [("mags", "float", (6))]
    result = np.zeros(len(spectra_list), dtype=dtype)

    filters = {}
    for filtername in keys:
        bp = np.loadtxt(
            os.path.join(through_path, "total_" + filtername + ".dat"),
            dtype=list(zip(["wave", "trans"], [float] * 2)),
        )
        temp_b = Bandpass()
        temp_b.set_bandpass(bp["wave"], bp["trans"])
        filters[filtername] = temp_b

    filterwave = np.array([filters[f].calc_eff_wavelen()[0] for f in keys])

    for i, spectrum in enumerate(spectra_list):
        tempSed = Sed()
        tempSed.set_sed(wave, flambda=spectrum)
        for j, filtName in enumerate(keys):
            try:
                result["mags"][i][j] = tempSed.calc_mag(filters[filtName])
            except:
                pass
    return result, filterwave


def recalc_mags(data_dir=None):
    """Recalculate the magnitudes for sky brightness components.

    DANGER:  Overwrites data files in place. The rubin_sim_data/skybrightness
    folder will need to be packaged and updated after running this to propigate
    changes ot rest of users.
    """
    dirs = ["Airglow", "MergedSpec", "ScatteredStarLight", "Zodiacal", "LowerAtm", "Moon", "UpperAtm"]

    if data_dir is None:
        data_dir = get_data_dir()

    full_paths = [os.path.join(data_dir, "skybrightness/ESO_Spectra", dirname) for dirname in dirs]
    for path in full_paths:
        files = glob.glob(os.path.join(path, "*.npz"))
        for filename in files:
            data = np.load(filename)

            spec = data["spec"].copy()
            wave = data["wave"].copy()
            data.close()
            new_mags, filterwave = spec2mags(spec["spectra"], wave)
            spec["mags"] = new_mags["mags"]

            np.savez(filename, wave=wave, spec=spec, filterWave=filterwave)

    pass



import unittest

import rubin_sim.maf.batches as batches


class TestCommon(unittest.TestCase):
    def test_col_map(self):
        colmap = batches.col_map_dict("opsimv4")
        self.assertEqual(colmap["raDecDeg"], True)
        self.assertEqual(colmap["ra"], "fieldRA")
        colmap = batches.col_map_dict("fbs")
        self.assertEqual(colmap["raDecDeg"], True)
        self.assertEqual(colmap["skyBrightness"], "skyBrightness")

    def test_filter_list(self):
        filterlist, colors, orders, sqls, info_label = batches.common.filter_list(all=False, extra_sql=None)
        self.assertEqual(len(filterlist), 6)
        self.assertEqual(len(colors), 6)
        self.assertEqual(sqls["u"], 'filter = "u"')
        filterlist, colors, orders, sqls, info_label = batches.common.filter_list(all=True, extra_sql=None)
        self.assertIn("all", filterlist)
        self.assertEqual(sqls["all"], "")
        filterlist, colors, orders, sqls, info_label = batches.common.filter_list(
            all=True, extra_sql="night=3"
        )
        self.assertEqual(sqls["all"], "night=3")
        self.assertEqual(sqls["u"], '(night=3) and (filter = "u")')
        self.assertEqual(info_label["u"], "night=3 u band")
        filterlist, colors, orders, sqls, info_label = batches.common.filter_list(
            all=True, extra_sql="night=3", extra_info_label="night 3"
        )
        self.assertEqual(info_label["u"], "night 3 u band")


if __name__ == "__main__":
    unittest.main()



# imports
import unittest

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.random import default_rng

from rubin_sim import maf

# constants

RANDOM_SEED = 6563

# exception classes

# interface functions

# classes


class TestSummaryPlots(unittest.TestCase):
    num_runs = 11
    num_metrics = 7

    def setUp(self):
        # Make tests reproducible
        self.rng = default_rng(RANDOM_SEED)

        self.runs = [f"run{i}" for i in range(self.num_runs)]
        self.baseline = self.runs[0]

        self.metrics = [f"metric{i}" for i in range(self.num_metrics)]
        self.inverted_metrics = ["metric3", "metric5", "metric6"]
        self.mag_metrics = ["metric1", "metric2", "metric6"]

        self.metric_values = pd.DataFrame(
            self.rng.normal(loc=3, scale=5, size=(self.num_runs, self.num_metrics)),
            columns=self.metrics,
            index=self.runs,
        )
        self.metric_values.columns.name = "metric"
        self.metric_values.index.name = "run"

        styles = ["-" for i in range(self.num_metrics)]

        self.metric_set = pd.DataFrame(
            {"mag": False, "invert": False, "metric": self.metrics, "style": styles}
        ).set_index("metric", drop=False)
        self.metric_set.loc[self.mag_metrics, "mag"] = True
        self.metric_set.loc[self.inverted_metrics, "invert"] = True
        self.metric_set.loc["metric3", "style"] = "b--"

    def test_normalize_metric_summaries(self):
        # Test standard normalization with one run
        norm_values = maf.normalize_metric_summaries(self.baseline, self.metric_values, self.metric_set)

        ref_norm_values = _run_infos_norm_df(
            self.metric_values,
            self.baseline,
            invert_cols=self.inverted_metrics,
            mag_cols=self.mag_metrics,
        )
        np.testing.assert_allclose(norm_values.values, ref_norm_values.values)

        # test normalizing against one run, as a list
        norm_values = maf.normalize_metric_summaries([self.baseline], self.metric_values, self.metric_set)
        np.testing.assert_allclose(norm_values.values, ref_norm_values.values)

        # test similar but pretend that self.baseline is two runs
        norm_values = maf.normalize_metric_summaries(
            [self.baseline, self.baseline], self.metric_values, self.metric_set
        )
        np.testing.assert_allclose(norm_values.values, ref_norm_values.values)

        # test similar but different runs
        norm_values = maf.normalize_metric_summaries(
            [self.runs[0], self.runs[1]], self.metric_values, self.metric_set
        )

    def test_plot_run_metric(self):
        fig, ax = maf.plot_run_metric(self.metric_values)

        fig, ax = maf.plot_run_metric(
            self.metric_values,
            baseline_run=self.baseline,
            metric_set=self.metric_set,
        )

        fig, ax = maf.plot_run_metric(
            self.metric_values,
            vertical_quantity="value",
            horizontal_quantity="run",
        )

        fig, ax = maf.plot_run_metric(
            self.metric_values,
            vertical_quantity="value",
            horizontal_quantity="metric",
        )

        run_label_map = {r: r + "foo" for r in self.runs}
        fig, ax = maf.plot_run_metric(self.metric_values, run_label_map=run_label_map)

        metric_label_map = {r: r + "foo" for r in self.metrics}
        fig, ax = maf.plot_run_metric(self.metric_values, metric_label_map=metric_label_map)

        fig, ax = maf.plot_run_metric(self.metric_values, cmap=plt.get_cmap("Set2"))

        fig, ax = maf.plot_run_metric(self.metric_values, linestyles=["-", "--", ":"])

        fig, ax = maf.plot_run_metric(self.metric_values, markers=["o", "v", "^", "<", ">", "8", "s"])

        fig, ax = plt.subplots()
        maf.plot_run_metric(self.metric_values, ax=ax)

    def test_plot_run_metric_mesh(self):
        fig, ax = maf.plot_run_metric_mesh(self.metric_values)

        fig, ax = maf.plot_run_metric_mesh(
            self.metric_values,
            baseline_run=self.baseline,
            metric_set=self.metric_set,
        )

        fig, ax = maf.plot_run_metric_mesh(
            self.metric_values,
            baseline_run=self.baseline,
        )

        fig, ax = maf.plot_run_metric_mesh(self.metric_values, color_range=12)

        run_label_map = {r: r + "foo" for r in self.runs}
        fig, ax = maf.plot_run_metric_mesh(self.metric_values, run_label_map=run_label_map)

        metric_label_map = {r: r + "foo" for r in self.metrics}
        fig, ax = maf.plot_run_metric_mesh(self.metric_values, metric_label_map=metric_label_map)

        fig, ax = maf.plot_run_metric_mesh(self.metric_values, cmap=plt.get_cmap("Spectral"))

        fig, ax = plt.subplots()
        maf.plot_run_metric_mesh(self.metric_values, ax=ax)


# internal functions & classes


def _run_infos_norm_df(df, norm_run, invert_cols=None, mag_cols=None):
    """
    Normalize values in a DataFrame, based on the values in a given run.
    Can normalize some columns (metric values) differently
    (invert_cols, reverse_cols, mag_cols)
    if those columns are specified; this lets the final normalized dataframe
    'look' the same way
    in a plot (i.e. "up" is better (reverse_cols), they center on 1 (mag_cols),
    and the magnitude scales as expected (invert_cols)).

    Parameters
    ----------
    df : pd.DataFrame
        The data frame containing the metric values to compare
    norm_run: str
        The name of the simulation to normalize to (typically family_baseline)
    invert_cols: list
        Columns (metric values) to convert to 1 / value
    mag_cols: list
        Columns (metrics values) to treat as magnitudes
        (1 + (difference from norm_run))

    Returns
    -------
    pd.DataFrame
        Normalized data frame
    """

    # Copy the dataframe but drop the columns containing only strings
    out_df = df.copy()
    if invert_cols is None:
        invert_cols = []
    if mag_cols is None:
        mag_cols = []
    both_cols = [c for c in invert_cols if c in mag_cols]
    invert_cols = [c for c in invert_cols if c not in both_cols]
    mag_cols = [c for c in mag_cols if c not in both_cols]
    out_df[mag_cols] = 1 + out_df[mag_cols] - out_df[mag_cols].loc[norm_run]
    out_df[invert_cols] = 1 / out_df[invert_cols]
    out_df[both_cols] = 1 - out_df[both_cols] + out_df[both_cols].loc[norm_run]
    # which columns are strings?
    string_cols = [c for c, t in zip(df.columns, df.dtypes) if t == "object"]
    cols = [c for c in out_df.columns.values if not (c in mag_cols or c in string_cols)]
    out_df[cols] = 1 + (out_df[cols] - out_df[cols].loc[norm_run]) / out_df[cols].loc[norm_run]
    return out_df


run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



__all__ = ("PeriodicDetectMetric",)

import numpy as np
import rubin_scheduler.utils as utils
import scipy

from rubin_sim.maf.utils import m52snr

from .base_metric import BaseMetric


class PeriodicDetectMetric(BaseMetric):
    """Determine if we would be able to classify an object as periodic/non-uniform, using an F-test
    The idea here is that if a periodic source is aliased, it will be indistinguishable from a constant source,
    so we can find a best-fit constant, and if the reduced chi-squared is ~1, we know we are aliased.

    Parameters
    ----------

    period : float (2) or array
        The period of the star (days). Can be a single value, or an array. If an array, amplitude and starMag
        should be arrays of equal length.
    amplitude : floar (0.1)
        The amplitude of the stellar variablility (mags).
    starMag : float (20.)
        The mean magnitude of the star in r (mags).
    sig_level : float (0.05)
        The value to use to compare to the p-value when deciding if we can reject the null hypothesis.
    sed_template : str ('F')
        The stellar SED template to use to generate realistic colors (default is an F star, so RR Lyrae-like)

    Returns
    -------

    1 if we would detect star is variable, 0 if it is well-fit by a constant value. If using arrays to test multiple
    period-amplitude-mag combinations, will be the sum of the number of detected stars.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        periods=2.0,
        amplitudes=0.1,
        m5_col="fiveSigmaDepth",
        metric_name="PeriodicDetectMetric",
        filter_col="filter",
        star_mags=20,
        sig_level=0.05,
        sed_template="F",
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        if np.size(periods) == 1:
            self.periods = [periods]
            # Using the same magnitude for all filters. Could expand to fit the mean in each filter.
            self.star_mags = [star_mags]
            self.amplitudes = [amplitudes]
        else:
            self.periods = periods
            self.star_mags = star_mags
            self.amplitudes = amplitudes
        self.sig_level = sig_level
        self.sed_template = sed_template

        super(PeriodicDetectMetric, self).__init__(
            [mjd_col, m5_col, filter_col],
            metric_name=metric_name,
            units="N Detected (0, %i)" % np.size(periods),
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        result = 0
        n_pts = np.size(data_slice[self.mjd_col])
        n_filt = np.size(np.unique(data_slice[self.filter_col]))

        # If we had a correct model with phase, amplitude, period, mean_mags,
        # then chi_squared/DoF would be ~1 with 3+n_filt free parameters.
        # The mean is one free parameter
        p1 = n_filt
        p2 = 3.0 + n_filt
        chi_sq_2 = 1.0 * (n_pts - p2)

        u_filters = np.unique(data_slice[self.filter_col])

        if n_pts > p2:
            for period, starMag, amplitude in zip(self.periods, self.star_mags, self.amplitudes):
                chi_sq_1 = 0
                mags = utils.stellar_mags(self.sed_template, rmag=starMag)
                for filtername in u_filters:
                    in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
                    lc = (
                        amplitude * np.sin(data_slice[self.mjd_col][in_filt] * (np.pi * 2) / period)
                        + mags[filtername]
                    )
                    snr = m52snr(lc, data_slice[self.m5_col][in_filt])
                    delta_m = 2.5 * np.log10(1.0 + 1.0 / snr)
                    weights = 1.0 / (delta_m**2)
                    weighted_mean = np.sum(weights * lc) / np.sum(weights)
                    chi_sq_1 += np.sum(((lc - weighted_mean) ** 2 / delta_m**2))
                # Yes, I'm fitting magnitudes rather than flux. At least I feel kinda bad about it.
                # F-test for nested models Regression problems:  https://en.wikipedia.org/wiki/F-test
                f_numerator = (chi_sq_1 - chi_sq_2) / (p2 - p1)
                f_denom = 1.0
                # This is just reduced chi-squared for the more complicated model, so should be 1.
                f_val = f_numerator / f_denom
                # Has DoF (p2-p1, n-p2)
                # https://stackoverflow.com/questions/21494141/how-do-i-do-a-f-test-in-python/21503346
                p_value = scipy.stats.f.sf(f_val, p2 - p1, n_pts - p2)
                if np.isfinite(p_value):
                    if p_value < self.sig_level:
                        result += 1

        return result



# imports
import sys
import unittest
from os import path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pandas as pd

from rubin_sim import maf

# constants

URLROOT = "https://raw.githubusercontent.com/lsst-pst/survey_strategy/main/fbs_2.0/"
FAMILY_SOURCE = URLROOT + "runs_v2.1.json"
METRIC_SET_SOURCE = URLROOT + "metric_sets.json"
SUMMARY_SOURCE = URLROOT + "summary_2022_04_28.csv"

# exception classes

# interface functions

# classes


class TestArchive(unittest.TestCase):
    num_runs = 11
    num_metrics = 7

    def test_get_family_runs(self):
        runs = maf.get_family_runs(FAMILY_SOURCE)

        self.assertIsInstance(runs, pd.DataFrame)

        columns = ("run", "brief", "filepath", "url")
        for column in columns:
            self.assertIn(column, runs.columns)

        self.assertEqual(runs.index.name, "family")

    def test_get_runs(self):
        runs = maf.get_runs(FAMILY_SOURCE)
        self.assertEqual(runs.index.name, "run")

        columns = ("family", "brief", "filepath", "url")
        for column in columns:
            self.assertIn(column, runs.columns)

        self.assertEqual(runs.index.name, "run")

    @unittest.skip("skipping long running test.")
    def test_download_runs(self):
        run = maf.get_runs(FAMILY_SOURCE).index.values[0]

        temp_dir_itself = TemporaryDirectory()
        temp_dir = temp_dir_itself.name

        dest_fnames = maf.download_runs(run, dest_dir=temp_dir)
        self.assertTrue(path.exists(dest_fnames[0]))

        temp_dir_itself.cleanup()

    def test_get_metric_sets(self):
        metric_sets = maf.get_metric_sets(METRIC_SET_SOURCE)
        self.assertIsInstance(metric_sets, pd.DataFrame)
        self.assertIn("metric set", metric_sets.index.names)
        self.assertIn("metric", metric_sets.index.names)

        columns = ("style", "invert", "mag")
        for column in columns:
            self.assertIn(column, metric_sets.columns)

    def test_get_metric_summaries(self):
        summary = maf.get_metric_summaries(summary_source=SUMMARY_SOURCE)
        self.assertIsInstance(summary, pd.DataFrame)
        self.assertEqual(summary.columns.name, "metric")
        self.assertEqual(summary.index.name, "run")

        my_summary = maf.get_metric_summaries(
            runs=["baseline_v2.0_10yrs", "baseline_retrofoot_v2.0_10yrs"],
            metrics=[
                "Rms Max normairmass All sky all bands HealpixSlicer",
                "Median Max normairmass All sky all bands HealpixSlicer",
                "Max Max normairmass All sky all bands HealpixSlicer",
            ],
            summary_source=summary,
        )
        self.assertEqual(len(my_summary), 2)
        self.assertEqual(len(my_summary.columns), 3)

        rolling_sum = maf.get_metric_summaries(
            run_families="rolling",
            summary_source=summary,
            runs_source=FAMILY_SOURCE,
        )
        self.assertGreater(len(rolling_sum), 3)
        self.assertLess(len(rolling_sum), len(summary))

        rollingddf_sum = maf.get_metric_summaries(
            run_families=["rolling", "ddf percent"],
            summary_source=summary,
            runs_source=FAMILY_SOURCE,
        )
        self.assertGreater(len(rollingddf_sum), len(rolling_sum))
        self.assertLess(len(rollingddf_sum), len(summary))

        srd_sim = maf.get_metric_summaries(
            metric_sets="SRD",
            summary_source=summary,
            metric_set_source=METRIC_SET_SOURCE,
        )
        self.assertGreater(len(srd_sim.columns), 3)
        self.assertLess(len(srd_sim.columns), len(summary.columns))

        srdnvis_sim = maf.get_metric_summaries(
            metric_sets=["SRD", "N Visits"],
            summary_source=summary,
            metric_set_source=METRIC_SET_SOURCE,
        )
        self.assertGreater(len(srdnvis_sim.columns), len(srd_sim.columns))
        self.assertLess(len(srdnvis_sim.columns), len(summary.columns))

    def test_get_family_descriptions(self):
        families = maf.get_family_descriptions(FAMILY_SOURCE)
        self.assertIsInstance(families, pd.DataFrame)
        self.assertEqual(families.index.name, "family")

    def test_describe_families(self):
        if "IPython" in sys.modules:
            with patch("IPython.display.display_markdown") as _, patch("IPython.display.HTML") as _, patch(
                "IPython.display"
            ) as _:
                self.perform_describe_families_test()
        else:
            with patch("builtins.print") as _:
                self.perform_describe_families_test()

    def perform_describe_families_test(self):
        families = maf.get_family_descriptions(FAMILY_SOURCE)
        disp_families = families[:2]

        fig, ax = maf.describe_families(families)
        self.assertIsNone(fig)
        self.assertIsNone(ax)

        all_metric_sets = maf.get_metric_sets(METRIC_SET_SOURCE)
        summary = maf.get_metric_summaries(summary_source=SUMMARY_SOURCE)

        table_metric_set = all_metric_sets.loc["SRD"]
        fig, ax = maf.describe_families(disp_families, summary=summary, table_metric_set=table_metric_set)
        self.assertIsNone(fig)
        self.assertIsNone(ax)

        plot_metric_set = all_metric_sets.loc["N Visits"]
        fig, ax = maf.describe_families(disp_families, summary=summary, plot_metric_set=plot_metric_set)

    def test_create_metric_set_df(self):
        metrics = ["Urania", "Thalia", "Calliope", "Terpsichore"]
        metric_set_name = "Muses"
        metric_set = maf.create_metric_set_df(metric_set_name, metrics)
        self.assertSequenceEqual(metrics, metric_set.metric.tolist())
        self.assertSequenceEqual(
            metric_set.columns.tolist(),
            ["metric", "short_name", "style", "invert", "mag"],
        )
        self.assertSequenceEqual(metric_set.index.names, ["metric set", "metric"])


run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



__all__ = (
    "combine_info_labels",
    "filter_list",
    "standard_summary",
    "extended_summary",
    "standard_metrics",
    "extended_metrics",
    "lightcurve_summary",
    "standard_angle_metrics",
    "summary_completeness_at_time",
    "summary_completeness_over_h",
    "fraction_population_at_threshold",
    "microlensing_summary",
)


import rubin_sim.maf.metrics as metrics


def combine_info_labels(info1, info2):
    if info1 is not None and info2 is not None:
        info = info1 + " " + info2
    elif info1 is not None:
        info = info1
    elif info2 is not None:
        info = info2
    else:
        info = None
    return info


def filter_list(all=True, extra_sql=None, extra_info_label=None):
    """Return a list of filters, plot colors and orders.

    Parameters
    ----------
    all : `bool`, optional
        Include 'all' in the list of filters and as part of the
        colors/orders dictionaries.
    extra_sql : `str`, optional
        Additional sql constraint to add to constraints returned per filter.
    extra_info_label : `str`, optional
        Substitute info_label to add to info_label strings composed per band.

    Returns
    -------
    filterlist : `list` [`str]
    colors : `dict` {`str`: `str`}
    orders : `dict` {`str`: int}
    sqls : `dict` {`str`: `str`}
    info_labels : `dict` {`str`: `str}
    """
    if all:
        filterlist = ("all", "u", "g", "r", "i", "z", "y")
    else:
        filterlist = ("u", "g", "r", "i", "z", "y")
    colors = {"u": "cyan", "g": "g", "r": "orange", "i": "r", "z": "m", "y": "b"}
    orders = {"u": 1, "g": 2, "r": 3, "i": 4, "z": 5, "y": 6}
    if all:
        colors["all"] = "k"
        orders["all"] = 0
    sqls = {}
    info_labels = {}
    if extra_info_label is None:
        if extra_sql is None or len(extra_sql) == 0:
            md = ""
        else:
            md = "%s " % extra_sql
    else:
        md = "%s " % extra_info_label
    for f in filterlist:
        if f == "all":
            sqls[f] = ""
            info_labels[f] = md + "all bands"
        else:
            sqls[f] = 'filter = "%s"' % f
            info_labels[f] = md + "%s band" % f
    if extra_sql is not None and len(extra_sql) > 0:
        for s in sqls:
            if s == "all":
                sqls[s] = extra_sql
            else:
                sqls[s] = "(%s) and (%s)" % (extra_sql, sqls[s])
    return filterlist, colors, orders, sqls, info_labels


def standard_summary(with_count=True):
    """A set of standard summary metrics, to calculate
    Mean, RMS, Median, #, Max/Min, and # 3-sigma outliers.

    Parameters
    ----------
    with_count : `bool`, optional
        Include the "Count" metric in the set of summary metrics or not.

    Returns
    -------
    standardSummary : `list` [`maf.BaseMetric`]
        List of metrics appropriate to use to summarize the results from
        another metric.
    """
    standardSummary = [
        metrics.MeanMetric(),
        metrics.RmsMetric(),
        metrics.MedianMetric(),
        metrics.MaxMetric(),
        metrics.MinMetric(),
        metrics.NoutliersNsigmaMetric(metric_name="N(+3Sigma)", n_sigma=3),
        metrics.NoutliersNsigmaMetric(metric_name="N(-3Sigma)", n_sigma=-3.0),
    ]
    if with_count:
        standardSummary += [metrics.CountMetric()]
    return standardSummary


def extended_summary():
    """An extended set of summary metrics, to calculate all that is in
    the standard summary stats, plus 25/75 percentiles.

    Returns
    --------
    extendedSummary : `list` [`maf.BaseMetric`]
        List of metrics appropriate to use to summarize the results
        from another metric.
    """
    extendedStats = standard_summary()
    extendedStats += [
        metrics.PercentileMetric(metric_name="25th%ile", percentile=25),
        metrics.PercentileMetric(metric_name="75th%ile", percentile=75),
    ]
    return extendedStats


def lightcurve_summary():
    lightcurveSummary = [
        metrics.SumMetric(metric_name="Total detected"),
        metrics.CountMetric(metric_name="Total lightcurves in footprint"),
        metrics.CountMetric(metric_name="Total lightcurves on sky", mask_val=0),
        metrics.MeanMetric(metric_name="Fraction detected in footprint (mean)"),
        metrics.MeanMetric(mask_val=0, metric_name="Fraction detected of total (mean)"),
    ]
    return lightcurveSummary


def standard_metrics(colname, replace_colname=None):
    """A set of standard simple metrics for some quantity.
    Typically would be applied with unislicer.

    Parameters
    ----------
    colname : `str`
        The column name to apply the metrics to.
    replace_colname: `str` or None, optional
        Value to replace colname with in the metric_name.
        i.e. if replace_colname='' then metric name is Mean,
        instead of Mean Airmass, or
        if replace_colname='seeingGeom', then metric name is
        Mean seeingGeom instead of Mean seeingFwhmGeom.

    Returns
    -------
    standardMetrics : `list` [`maf.BaseMetric`]
        List of appropriate MAF metrics to evaluate a distribution.
    """
    standardMetrics = [
        metrics.MeanMetric(colname),
        metrics.MedianMetric(colname),
        metrics.MinMetric(colname),
        metrics.MaxMetric(colname),
    ]
    if replace_colname is not None:
        for m in standardMetrics:
            if len(replace_colname) > 0:
                m.name = m.name.replace("%s" % colname, "%s" % replace_colname)
            else:
                m.name = m.name.rstrip(" %s" % colname)
    return standardMetrics


def extended_metrics(colname, replace_colname=None):
    """An extended set of simple metrics for some quantity.
    Typically applied with unislicer.

    Parameters
    ----------
    colname : `str`
        The column name to apply the metrics to.
    replace_colname: `str` or None, optional
        Value to replace colname with in the metric_name.
        i.e. if replace_colname='' then metric name is Mean,
        instead of Mean Airmass, or
        if replace_colname='seeingGeom', then metric name is
        Mean seeingGeom instead of Mean seeingFwhmGeom.

    Returns
    -------
    extendedMetrics : `list` [`maf.BaseMetric`]
        List of appropriate MAF metrics to evaluate a distribution.
    """
    extendedMetrics = standard_metrics(colname, replace_colname=None)
    extendedMetrics += [
        metrics.RmsMetric(colname),
        metrics.NoutliersNsigmaMetric(colname, metric_name="N(+3Sigma) " + colname, n_sigma=3),
        metrics.NoutliersNsigmaMetric(colname, metric_name="N(-3Sigma) " + colname, n_sigma=-3),
        metrics.PercentileMetric(colname, percentile=25),
        metrics.PercentileMetric(colname, percentile=75),
        metrics.CountMetric(colname),
    ]
    if replace_colname is not None:
        for m in extendedMetrics:
            if len(replace_colname) > 0:
                m.name = m.name.replace("%s" % colname, "%s" % replace_colname)
            else:
                m.name = m.name.rstrip(" %s" % colname)
    return extendedMetrics


def standard_angle_metrics(colname, replace_colname=None):
    """A set of standard simple metrics for a wrap-around angle quantity.

    Parameters
    ----------
    colname : `str`
        The column name to apply the metrics to.
    replace_colname: `str` or None, optional
        Value to replace colname with in the metric_name.
        i.e. if replace_colname='' then metric name is Mean,
        instead of Mean Airmass, or
        if replace_colname='seeingGeom', then metric name is
        Mean seeingGeom instead of Mean seeingFwhmGeom.

    Returns
    -------
    standardAngleMetrics : `list` [`maf.BaseMetric`]
        List of appropriate MAF metrics for angle distributions.
    """
    standardAngleMetrics = [
        metrics.MeanAngleMetric(colname),
        metrics.RmsAngleMetric(colname),
        metrics.FullRangeAngleMetric(colname),
        metrics.MinMetric(colname),
        metrics.MaxMetric(colname),
    ]
    if replace_colname is not None:
        for m in standardAngleMetrics:
            if len(replace_colname) > 0:
                m.name = m.name.replace("%s" % colname, "%s" % replace_colname)
            else:
                m.name = m.name.rstrip(" %s" % colname)
    return standardAngleMetrics


def summary_completeness_at_time(times, h_val, h_index=0.33):
    """A simple list of summary metrics to be applied to the Discovery_Time
    or PreviouslyKnown metrics.
    (can be used with any moving object metric which returns the
    time of discovery).

    Parameters
    ----------
    times : `np.ndarray `or list` [`float`]
        The times at which to evaluate the completeness @ Hval.
    h_val : `float`
        The H value at which to evaluate the completeness
        (cumulative and differential).
    h_index : `float`, optional
        The index of the power law to integrate H over
        (for cumulative completeness).

    Returns
    -------
    summaryMetrics : `list` [`maf.MoCompletenessAtTimeMetric`]
        List of completeness metrics to be evaluated at the specified times.
    """
    summaryMetrics = [
        metrics.MoCompletenessAtTimeMetric(times=times, hval=h_val, hindex=h_index, cumulative=False),
        metrics.MoCompletenessAtTimeMetric(times=times, hval=h_val, hindex=h_index, cumulative=True),
    ]
    return summaryMetrics


def summary_completeness_over_h(requiredChances=1, Hindex=0.33):
    """A simple list of summary metrics to be applied to the
    Discovery_N_Chances metric.

    Parameters
    ----------
    requiredChances : `int`, optional
        Number of discovery opportunities required to consider an
        object 'discovered'.
    Hindex : `float`, optional
        The index of the power law to integrate H over
        (for cumulative completeness).

    Returns
    -------
    summaryMetrics : `list` [`maf.MoCompletenessMetric`]
        List of moving object MoCompleteness metrics
        (cumulative and differential)
    """
    summaryMetrics = [
        metrics.MoCompletenessMetric(threshold=requiredChances, cumulative=False, hindex=Hindex),
        metrics.MoCompletenessMetric(threshold=requiredChances, cumulative=True, hindex=Hindex),
    ]
    return summaryMetrics


def fraction_population_at_threshold(thresholds, optnames=None):
    """Creates a list of summary metrics to be applied to any moving object
    metric which reports a float value, calculating the fraction of the
    population above X.

    Parameters
    ----------
    thresholds : `list` [`float`]
        The thresholds at which to calculate what fraction of the population
        exceeds these values.
    optnames : `list` [`str`], optional
        If provided, these names will be used instead of the threshold values
        when constructing the metric names.
        This allows more descriptive summary statistic names.

    Returns
    -------
    fracMetrics : `list` [`maf.MoCompletenessMetric`]
        List of moving object MoCompleteness metrics
        (differential fractions of the population).
    """
    fracMetrics = []
    for i, threshold in enumerate(thresholds):
        if optnames is not None:
            o = optnames[i]
        else:
            o = threshold
        m = metrics.MoCompletenessMetric(
            threshold=threshold, cumulative=False, metric_name=f"FractionPop {o}"
        )
        fracMetrics.append(m)
    return fracMetrics


def microlensing_summary(metric_type, npts_required=10, Fisher_sigmatE_tE_cutoff=0.1):
    """Calculate summary metrics for the microlensing population metrics.

    Parameters
    -----------
    metric_type : `str`
        Identify whether the metric is "Npts" or Fisher"
    npts_required : `int`, optional
        Count the fraction of microlensing events with more than npts_required
        observations
    Fisher_sigmatE_tE_cutoff : `float`, optional
        Count the fraction of microlensing events with characterization
        uncertainty less than this

    Returns
    -------
    microlensingSummary : `list` [`maf.BaseMetric`]
        List of appropriate MAF metrics for this type of microlensing
        metric with the specified threshold values.
    """
    if metric_type != "Npts" and metric_type != "Fisher":
        raise Exception('metric_type must be "Npts" or "Fisher"')
    if metric_type == "Npts":
        microlensingSummary = [
            metrics.FracAboveMetric(
                cutoff=npts_required,
                metric_name=f"Fraction w/ at least {npts_required} points",
            ),
            metrics.CountMetric(metric_name="Total lightcurves in footprint"),
            metrics.CountMetric(metric_name="Total lightcurves on sky", mask_val=0),
            metrics.MeanMetric(metric_name="Mean number of points per lightcurves in footprint"),
            metrics.MeanMetric(mask_val=0, metric_name="Mean number of points per lightcurves in total"),
        ]
    elif metric_type == "Fisher":
        microlensingSummary = [
            metrics.FracBelowMetric(
                cutoff=Fisher_sigmatE_tE_cutoff,
                metric_name=f"Fraction w/ sigma_tE/tE < {Fisher_sigmatE_tE_cutoff}",
            ),
            metrics.CountMetric(metric_name="Total lightcurves in footprint"),
            metrics.CountMetric(metric_name="Total lightcurves on sky", mask_val=0),
            metrics.RealMeanMetric(metric_name="Mean sigma_tE/tE in footprint (mean)"),
            metrics.RealMeanMetric(mask_val=0, metric_name="Mean sigma_tE/tE of total (mean)"),
        ]
    return microlensingSummary



__all__ = ("StarDensityMetric",)

import warnings

from scipy.interpolate import interp1d

from .base_metric import BaseMetric


class StarDensityMetric(BaseMetric):
    """Interpolate the stellar luminosity function to return the number of
    stars per square arcsecond brighter than the mag_limit.
    Note that the map is built from CatSim stars in the range 20 < r < 28.
    mag_limit values outside that the range of the map's starMapBins will return self.badval

    The stellar density maps are available in any bandpass, but bandpasses other
    than r band must use a pre-configured StellarDensityMap (not just the default).
    In other words, when setting up the metric bundle for an i-band stellar density
    using (as an example) a HealpixSlicer:
    ```
    map = maf.StellarDensityMap(filtername='i')
    metric = maf.StarDensityMetric(filtername='i', mag_limit=25.0)
    slicer = maf.HealpixSlicer()
    bundle = maf.MetricBundle(metric, slicer, "", mapsList=[map])
    ```

    Parameters
    ----------
    mag_limit : `float`, opt
        Magnitude limit at which to evaluate the stellar luminosity function.
        Returns number of stars per square arcsecond brighter than this limit.
        Default 25.
    filtername : `str`, opt
        Which filter to evaluate the luminosity function in; Note that using bands other than r
        will require setting up a custom (rather than default) version of the stellar density map.
        Default r.
    units : `str`, opt
        Units for the output values. Default "stars/sq arcsec".
    maps : `list` of `str`, opt
        Names for the maps required. Default "StellarDensityMap".

    Returns
    -------
    result : `float`
        Number of stars brighter than mag_limit in filtername, based on the stellar density map.
    """

    def __init__(
        self, mag_limit=25.0, filtername="r", units="stars/sq arcsec", maps=["StellarDensityMap"], **kwargs
    ):
        super(StarDensityMetric, self).__init__(col=[], maps=maps, units=units, **kwargs)
        self.mag_limit = mag_limit
        if "rmagLimit" in kwargs:
            warnings.warn(
                "rmagLimit is deprecated; please use mag_limit instead "
                "(will use the provided rmagLimit for now)."
            )
            self.mag_limit = kwargs["rmagLimit"]
        self.filtername = filtername

    def run(self, data_slice, slice_point=None):
        # Interpolate the data to the requested mag
        interp = interp1d(
            slice_point["starMapBins_%s" % self.filtername][1:],
            slice_point["starLumFunc_%s" % self.filtername],
        )
        # convert from stars/sq degree to stars/sq arcsec
        try:
            result = interp(self.mag_limit) / (3600.0**2)
        except ValueError:
            # This probably means the interpolation went out of range (magLimit <15 or >28)
            return self.badval
        return result



__all__ = ("info_bundle_dicts",)

import numpy as np

import rubin_sim.maf.batches as batches


def info_bundle_dicts(allsky_slicer, wfd_slicer, opsim="opsim", colmap=batches.col_map_dict()):
    # Set up the bundle dicts
    # Some of these metrics are reproduced in other scripts - srd and cadence
    bdict = {}

    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        fO = batches.fOBatch(colmap=colmap, runName=opsim, slicer=slicer, extraInfoLabel=tag)
        bdict.update(fO)
        astrometry = batches.astrometryBatch(colmap=colmap, runName=opsim, slicer=slicer, extraInfoLabel=tag)
        bdict.update(astrometry)
        rapidrevisit = batches.rapidRevisitBatch(
            colmap=colmap, runName=opsim, slicer=slicer, extraInfoLabel=tag
        )
        bdict.update(rapidrevisit)

    # Intranight (pairs/time)
    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        intranight = batches.intraNight(colmap, opsim, slicer=slicer, extraInfoLabel=tag)
        bdict.update(intranight)

    # Internight (nights between visits)
    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        internight = batches.interNight(colmap, opsim, slicer=slicer, extraInfoLabel=tag)
        bdict.update(internight)

    # Intraseason (length of season)
    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        season = batches.seasons(colmap=colmap, runName=opsim, slicer=slicer, extraInfoLabel=tag)
        bdict.update(season)

    # Run all metadata metrics, All and just WFD.
    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        bdict.update(batches.allMetadata(colmap, opsim, slicer=slicer, extraInfoLabel=tag))

    # And run some metadata for the first year only - all sky
    bdict.update(batches.firstYearMetadata(colmap, opsim, slicer=allsky_slicer))

    # Nvisits + m5 maps + Teff maps, All and just WFD.
    for tag, slicer in zip(["All sky", "WFD"], [allsky_slicer, wfd_slicer]):
        bdict.update(batches.nvisitsM5Maps(colmap, opsim, slicer=slicer, extraInfoLabel=tag))
        bdict.update(batches.tEffMetrics(colmap, opsim, slicer=slicer, extraInfoLabel=tag))

    # And number of visits for the first year and halfway through the survey
    bdict.update(
        batches.nvisitsM5Maps(
            colmap,
            opsim,
            slicer=allsky_slicer,
            extraInfoLabel="Yr 1",
            extraSql="night < 365.5",
            runLength=1,
        )
    )
    bdict.update(
        batches.nvisitsM5Maps(
            colmap,
            opsim,
            slicer=allsky_slicer,
            extraSql='night > 365*3.5 and night < 365*4.5 and note not like "%DD%"',
            extraInfoLabel="Yr 3-4",
            runLength=1,
        )
    )

    # NVisits alt/az LambertSkyMap (all filters, per filter)
    bdict.update(batches.altazLambert(colmap, opsim))

    # Slew metrics.
    bdict.update(batches.slewBasics(colmap, opsim))

    # Open shutter metrics.
    bdict.update(batches.openshutterFractions(colmap, opsim))

    # Some basic nvisits per purpose metrics
    wfd_footprint_mask = np.where(wfd_slicer.mask, 0, 1)
    bdict.update(
        batches.nvisitsPerSubset(
            colmap,
            opsim,
            constraint="visitExposureTime > 19",
            footprintConstraint=wfd_footprint_mask,
            extraInfoLabel="WFD",
        )
    )
    dd_constraint = "note like '%DD%'"
    bdict.update(
        batches.nvisitsPerSubset(
            colmap,
            opsim,
            constraint=dd_constraint,
            footprintConstraint=None,
            extraInfoLabel="DDF",
        )
    )
    bdict.update(
        batches.nvisitsPerSubset(
            colmap,
            opsim,
            constraint=None,
            footprintConstraint=None,
            extraInfoLabel="All visits",
        )
    )

    # Per night and whole survey filter changes.
    bdict.update(batches.filtersPerNight(colmap, opsim, nights=1, extraSql=None))
    bdict.update(batches.filtersWholeSurvey(colmap, opsim, extraSql=None))

    return bdict



import matplotlib

matplotlib.use("Agg")

import unittest

import numpy as np

import rubin_sim.maf.metric_bundles as metricBundle
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers


class Test2D(unittest.TestCase):
    def setUp(self):
        names = [
            "night",
            "fieldId",
            "fieldRA",
            "fieldDec",
            "fiveSigmaDepth",
            "observationStartMJD",
            "rotSkyPos",
        ]
        types = [int, int, float, float, float, float, float]

        self.m5_1 = 25.0
        self.m5_2 = 24.0

        self.n1 = 50
        self.n2 = 49

        # Picking RA and Dec values that will hit nside=16 healpixels
        self.sim_data = np.zeros(self.n1 + self.n2, dtype=list(zip(names, types)))
        self.sim_data["night"][0 : self.n1] = 1
        self.sim_data["fieldId"][0 : self.n1] = 1
        self.sim_data["fieldRA"][0 : self.n1] = 10.0
        self.sim_data["fieldDec"][0 : self.n1] = 0.0
        self.sim_data["fiveSigmaDepth"][0 : self.n1] = self.m5_1

        self.sim_data["night"][self.n1 :] = 2
        self.sim_data["fieldId"][self.n1 :] = 2
        self.sim_data["fieldRA"][self.n1 :] = 190.0
        self.sim_data["fieldDec"][self.n1 :] = -20.0
        self.sim_data["fiveSigmaDepth"][self.n1 :] = self.m5_2

        self.field_data = np.zeros(
            2, dtype=list(zip(["fieldId", "fieldRA", "fieldDec"], [int, float, float]))
        )
        self.field_data["fieldId"] = [1, 2]
        self.field_data["fieldRA"] = np.radians([10.0, 190.0])
        self.field_data["fieldDec"] = np.radians([0.0, -20.0])

        self.sim_data["observationStartMJD"] = self.sim_data["night"]

    def test_user_points2d_slicer(self):
        metric = metrics.AccumulateCountMetric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.UserPointsSlicer(
            ra=np.degrees(self.field_data["fieldRA"]),
            dec=np.degrees(self.field_data["fieldDec"]),
            lat_lon_deg=True,
        )
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.fieldData = self.field_data
        mbg.run_current("", sim_data=self.sim_data)
        expected = np.array([[self.n1, self.n1], [-666.0, self.n2]])
        assert np.array_equal(mb.metric_values.data, expected)

    def test_healpix2d_slicer(self):
        metric = metrics.AccumulateCountMetric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)

        good = np.where(mb.metric_values.mask[:, -1] == False)[0]
        expected = np.array([[self.n1, self.n1], [-666.0, self.n2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

    def test_histogram_metric(self):
        metric = metrics.HistogramMetric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)

        good = np.where(mb.metric_values.mask[:, -1] == False)[0]
        expected = np.array([[self.n1, 0.0], [0.0, self.n2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

        # Check that I can run a different statistic
        metric = metrics.HistogramMetric(col="fiveSigmaDepth", statistic="sum", bins=[0.5, 1.5, 2.5])
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)
        expected = np.array([[self.m5_1 * self.n1, 0.0], [0.0, self.m5_2 * self.n2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

    def test_accumulate_metric(self):
        metric = metrics.AccumulateMetric(col="fiveSigmaDepth", bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)
        good = np.where(mb.metric_values.mask[:, -1] == False)[0]
        expected = np.array([[self.n1 * self.m5_1, self.n1 * self.m5_1], [-666.0, self.n2 * self.m5_2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

    def test_histogram_m5_metric(self):
        metric = metrics.HistogramM5Metric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)
        good = np.where((mb.metric_values.mask[:, 0] == False) | (mb.metric_values.mask[:, 1] == False))[0]

        check_metric = metrics.Coaddm5Metric()
        temp_slice = np.zeros(self.n1, dtype=list(zip(["fiveSigmaDepth"], [float])))
        temp_slice["fiveSigmaDepth"] += self.m5_1
        val1 = check_metric.run(temp_slice)
        temp_slice = np.zeros(self.n2, dtype=list(zip(["fiveSigmaDepth"], [float])))
        temp_slice["fiveSigmaDepth"] += self.m5_2
        val2 = check_metric.run(temp_slice)

        expected = np.array([[val1, -666.0], [-666.0, val2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

    def test_accumulate_m5_metric(self):
        metric = metrics.AccumulateM5Metric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        mb = metricBundle.MetricBundle(metric, slicer, sql)
        # Clobber the stacker that gets auto-added
        mb.stacker_list = []
        mbg = metricBundle.MetricBundleGroup({0: mb}, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)
        good = np.where(mb.metric_values.mask[:, -1] == False)[0]

        check_metric = metrics.Coaddm5Metric()
        temp_slice = np.zeros(self.n1, dtype=list(zip(["fiveSigmaDepth"], [float])))
        temp_slice["fiveSigmaDepth"] += self.m5_1
        val1 = check_metric.run(temp_slice)
        temp_slice = np.zeros(self.n2, dtype=list(zip(["fiveSigmaDepth"], [float])))
        temp_slice["fiveSigmaDepth"] += self.m5_2
        val2 = check_metric.run(temp_slice)

        expected = np.array([[val1, val1], [-666.0, val2]])
        assert np.array_equal(mb.metric_values.data[good, :], expected)

    def test_accumulate_uniformity_metric(self):
        names = ["night"]
        types = ["float"]
        data_slice = np.zeros(3652, dtype=list(zip(names, types)))

        # Test that a uniform distribution is very close to zero
        data_slice["night"] = np.arange(1, data_slice.size + 1)
        metric = metrics.AccumulateUniformityMetric()
        result = metric.run(data_slice)
        assert np.max(result) < 1.0 / 365.25
        assert np.min(result) >= 0

        # Test that if everythin on night 1 or last night, then result is ~1
        data_slice["night"] = 1
        result = metric.run(data_slice)
        assert np.max(result) >= 1.0 - 1.0 / 365.25
        data_slice["night"] = 3652
        result = metric.run(data_slice)
        assert np.max(result) >= 1.0 - 1.0 / 365.25

        # Test if all taken in the middle, result ~0.5
        data_slice["night"] = 3652 / 2
        result = metric.run(data_slice)
        assert np.max(result) >= 0.5 - 1.0 / 365.25

    def test_run_regular_too(self):
        """
        Test that a binned slicer and a regular slicer can run together
        """
        bundle_list = []
        metric = metrics.AccumulateM5Metric(bins=[0.5, 1.5, 2.5])
        slicer = slicers.HealpixSlicer(nside=16)
        sql = ""
        bundle_list.append(metricBundle.MetricBundle(metric, slicer, sql))
        metric = metrics.Coaddm5Metric()
        slicer = slicers.HealpixSlicer(nside=16)
        bundle_list.append(metricBundle.MetricBundle(metric, slicer, sql))
        for bundle in bundle_list:
            bundle.stacker_list = []
        bd = metricBundle.make_bundles_dict_from_list(bundle_list)
        mbg = metricBundle.MetricBundleGroup(bd, None, save_early=False)
        mbg.set_current("")
        mbg.run_current("", sim_data=self.sim_data)

        assert np.array_equal(
            bundle_list[0].metric_values[:, 1].compressed(),
            bundle_list[1].metric_values.compressed(),
        )


if __name__ == "__main__":
    unittest.main()



__all__ = ("run_moving_calc",)

import argparse
import os

import numpy as np

from rubin_sim.maf.slicers import MoObjSlicer
from rubin_sim.utils import survey_start_mjd

from . import batches as batches
from . import db as db
from . import metricBundles as mmb


def run_moving_calc():
    """Calculate metric values for an input population. Can be used on either
    a split or complete population. If running on a split population for later
    re-combining, use the complete set of orbits as the 'orbit_file'. Assumes
    you have already created the moving object observation files.
    """

    parser = argparse.ArgumentParser(description="Run moving object metrics for a particular scheduler run.")
    parser.add_argument("--orbit_file", type=str, help="File containing the moving object orbits.")
    parser.add_argument(
        "--objtype",
        type=str,
        default="",
        help="Object type, to set up default characterization/H value parameters "
        "(if they're not specified by flags) and to label outputs.",
    )
    parser.add_argument(
        "--obs_file",
        type=str,
        help="File containing the observations of the moving objects.",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default=".",
        help="Output directory for moving object metrics. Default '.'.",
    )
    parser.add_argument(
        "--simulation_db",
        type=str,
        default=None,
        help="Path and filename of opsim output sqlite file, to write config* files to output directory."
        " Optional: if not provided, files won't have useful names.",
    )
    parser.add_argument(
        "--h_min",
        type=float,
        help="Minimum H value. " "If not set, defaults from objtype will be used.",
    )
    parser.add_argument(
        "--h_max",
        type=float,
        help="Maximum H value. " "If not set, defaults from objtype will be used.",
    )
    parser.add_argument(
        "--h_step",
        type=float,
        help="Stepsizes in H values. " "If not set, defaults from objtype will be used.",
    )
    parser.add_argument(
        "--h_mark",
        type=float,
        default=None,
        help="Add vertical lines at H=h_mark on plots. Default None.",
    )
    parser.add_argument(
        "--characterization",
        type=str,
        help="Inner/Outer solar system characterization. " "If unset, defaults from objtype will be used.",
    )
    parser.add_argument(
        "--constraint_info_label",
        type=str,
        default="",
        help="Metadata to add to the output files beyond objtype. Typically translation of "
        "the sql constraint into something more readable.",
    )
    parser.add_argument(
        "--constraint",
        type=str,
        default=None,
        help="(sql-style) constraint to apply to the solar system observations.",
    )
    parser.add_argument(
        "--albedo",
        type=float,
        default=None,
        help="Albedo value, to add diameters to upper scales on plots. Default None.",
    )

    parser.add_argument(
        "--n_years_max",
        type=int,
        default=10,
        help="Maximum number of years out to which to evaluate completeness." "Default 10.",
    )
    parser.add_argument(
        "--start_time",
        type=float,
        default=None,
        help="Time at start of survey (to set time for summary metrics).",
    )
    args = parser.parse_args()

    run_name = os.path.split(args.simulation_db)[-1].replace(".db", "")

    if args.orbit_file is None or args.obs_file is None:
        print("Must specify an orbit_file and an obs_file to calculate the metrics.")
        exit()

    # Get default H and other values:
    defaults = batches.ss_population_defaults(args.objtype)
    h_min = defaults["h_range"][0]
    h_max = defaults["h_range"][1]
    h_step = defaults["h_range"][2]
    h_mark = defaults["h_mark"]
    characterization = defaults["char"]
    magtype = defaults["magtype"]
    # If H info is specified from parser, use that
    if args.h_min is not None:
        h_min = args.h_min
    if args.h_max is not None:
        h_max = args.h_max
    if args.h_step is not None:
        h_step = args.h_step
    if args.h_mark is not None:
        h_mark = args.h_mark
    if args.characterization is not None:
        characterization = args.characterization
    Hrange = np.arange(h_min, h_max + h_step, h_step)

    # Default parameters for metric setup.
    if args.start_time is None:
        start_time = survey_start_mjd()
    else:
        start_time = args.start_time

    stepsize = 365 / 2.0
    times = np.arange(0, args.n_years_max * 365 + stepsize / 2, stepsize)
    times += start_time

    # Set up results_db.
    if not (os.path.isdir(args.out_dir)):
        try:
            os.makedirs(args.out_dir)
        except FileExistsError:
            # This can happen if you are running these in parallel and two scripts try to make
            # the same directory.
            pass
    results_db = db.ResultsDb(out_dir=args.out_dir)

    colmap = batches.col_map_dict()
    slicer = MoObjSlicer(h_range=Hrange)
    slicer.setup_slicer(orbit_file=args.orbit_file, obs_file=args.obs_file)
    # Run discovery metrics using 'trailing' losses
    bdictT = batches.quick_discovery_batch(
        slicer,
        colmap=colmap,
        run_name=run_name,
        objtype=args.objtype,
        constraint_info_label=args.constraint_info_label,
        constraint=args.constraint,
        detection_losses="trailing",
        albedo=args.albedo,
        h_mark=h_mark,
    )
    # Run these discovery metrics
    print("Calculating quick discovery metrics with simple trailing losses.")
    bg = mmb.MoMetricBundleGroup(bdictT, out_dir=args.out_dir, results_db=results_db)
    bg.run_all()

    # Run all discovery metrics using 'detection' losses
    bdictD = batches.quick_discovery_batch(
        slicer,
        colmap=colmap,
        run_name=run_name,
        objtype=args.objtype,
        constraint_info_label=args.constraint_info_label,
        constraint=args.constraint,
        detection_losses="detection",
        albedo=args.albedo,
        h_mark=h_mark,
    )
    bdict = batches.discovery_batch(
        slicer,
        colmap=colmap,
        run_name=run_name,
        objtype=args.objtype,
        constraint_info_label=args.constraint_info_label,
        constraint=args.constraint,
        detection_losses="detection",
        albedo=args.albedo,
        h_mark=h_mark,
    )
    bdictD.update(bdict)

    # Run these discovery metrics
    print("Calculating full discovery metrics with detection losses.")
    bg = mmb.MoMetricBundleGroup(bdictD, out_dir=args.out_dir, results_db=results_db)
    bg.run_all()

    # Run all characterization metrics
    if characterization.lower() == "inner":
        bdictC = batches.characterization_inner_batch(
            slicer,
            colmap=colmap,
            run_name=run_name,
            objtype=args.objtype,
            albedo=args.albedo,
            constraint_info_label=args.constraint_info_label,
            constraint=args.constraint,
            h_mark=h_mark,
        )
    elif characterization.lower() == "outer":
        bdictC = batches.characterization_outer_batch(
            slicer,
            colmap=colmap,
            run_name=run_name,
            objtype=args.objtype,
            albedo=args.albedo,
            constraint_info_label=args.constraint_info_label,
            constraint=args.constraint,
            h_mark=h_mark,
        )
    # Run these characterization metrics
    print("Calculating characterization metrics.")
    bg = mmb.MoMetricBundleGroup(bdictC, out_dir=args.out_dir, results_db=results_db)
    bg.run_all()



from .base_map import *
from .dust_map import *
from .dust_map_3d import *
from .ebv_3d_hp import *
from .gal_coords_map import *
from .galactic_plane_priority_maps import *
from .stellar_density_map import *
from .trilegal_map import *



__all__ = ("MoObjSlicer",)

import numpy as np
import pandas as pd

from rubin_sim.maf.plots.mo_plotters import MetricVsH, MetricVsOrbit
from rubin_sim.moving_objects.orbits import Orbits

from .base_slicer import BaseSlicer


class MoObjSlicer(BaseSlicer):
    """Slice moving object _observations_, per object and optionally clone/per H value.

    Iteration over the MoObjSlicer will go as:
    * iterate over each orbit;
    * if Hrange is not None, for each orbit, iterate over Hrange.

    Parameters
    ----------
    h_range : numpy.ndarray or None
        The H values to clone the orbital parameters over. If Hrange is None, will not clone orbits.
    """

    def __init__(self, h_range=None, verbose=True, badval=0):
        super(MoObjSlicer, self).__init__(verbose=verbose, badval=badval)
        self.Hrange = h_range
        self.slicer_init = {"h_range": h_range, "badval": badval}
        # Set default plot_funcs.
        self.plot_funcs = [
            MetricVsH(),
            MetricVsOrbit(xaxis="q", yaxis="e"),
            MetricVsOrbit(xaxis="q", yaxis="inc"),
        ]

    def setup_slicer(self, orbit_file, delim=None, skiprows=None, obs_file=None):
        """Set up the slicer and read orbit_file and obs_file from disk.

        Sets self.orbits (with orbit parameters), self.all_obs, and self.obs
        self.orbit_file and self.obs_file

        Parameters
        ----------
        orbit_file : str
            The file containing the orbit information.
            This is necessary, in order to be able to generate plots.
        obs_file : str, optional
            The file containing the observations of each object, optional.
            If not provided (default, None), then the slicer will not be able to 'slice', but can still plot.
        """
        self.read_orbits(orbit_file, delim=delim, skiprows=skiprows)
        if obs_file is not None:
            self.read_obs(obs_file)
        else:
            self.obs_file = None
            self.all_obs = None
            self.obs = None
        # Add these filenames to the slicer init values, to preserve in output files.
        self.slicer_init["orbit_file"] = self.orbit_file
        self.slicer_init["obs_file"] = self.obs_file

    def read_orbits(self, orbit_file, delim=None, skiprows=None):
        # Use sims_movingObjects to read orbit files.
        orb = Orbits()
        orb.read_orbits(orbit_file, delim=delim, skiprows=skiprows)
        self.orbit_file = orbit_file
        self.orbits = orb.orbits
        # Then go on as previously. Need to refactor this into 'setup_slicer' style.
        self.nSso = len(self.orbits)
        self.slice_points = {}
        self.slice_points["orbits"] = self.orbits
        # And set the slicer shape/size.
        if self.Hrange is not None:
            self.shape = [self.nSso, len(self.Hrange)]
            self.slice_points["H"] = self.Hrange
        else:
            self.shape = [self.nSso, 1]
            self.slice_points["H"] = self.orbits["H"]
        # Set the rest of the slice_point information once
        self.nslice = self.shape[0] * self.shape[1]

    def read_obs(self, obs_file):
        """Read observations of the solar system objects (such as created by sims_movingObjects).

        Parameters
        ----------
        obs_file: str
            The file containing the observation information.
        """
        # For now, just read all the observations (should be able to chunk this though).
        restore_file = np.load(obs_file)
        self.all_obs = restore_file["object_observations"].copy()
        restore_file.close()
        self.obs_file = obs_file
        self.all_obs = pd.DataFrame(self.all_obs)

        if "velocity" not in self.all_obs.columns:
            self.all_obs["velocity"] = np.sqrt(self.all_obs["dradt"] ** 2 + self.all_obs["ddecdt"] ** 2)

    def subset_obs(self, pandas_constraint=None):
        """
        Choose a subset of all the observations, such as those in a particular time period.
        """
        if pandas_constraint is None:
            self.obs = self.all_obs
        else:
            self.obs = self.all_obs.query(pandas_constraint)

    def _slice_obs(self, idx):
        """Return the observations of a given ssoId.

        For now this works for any ssoId; in the future, this might only work as ssoId is
        progressively iterated through the series of ssoIds (so we can 'chunk' the reading).

        Parameters
        ----------
        idx : integer
            The integer index of the particular SSO in the orbits dataframe.
        """
        # Find the matching orbit.
        orb = self.orbits.iloc[idx]
        # Find the matching observations.
        if self.obs["obj_id"].dtype == "object":
            obs = self.obs.query('obj_id == "%s"' % (orb["obj_id"]))
        else:
            obs = self.obs.query("obj_id == %d" % (orb["obj_id"]))
        # Return the values for H to consider for metric.
        if self.Hrange is not None:
            Hvals = self.Hrange
        else:
            Hvals = np.array([orb["H"]], float)
        # Note that ssoObs / obs is a recarray not Dataframe!
        # But that the orbit IS a Dataframe.
        return {"obs": obs.to_records(), "orbit": orb, "Hvals": Hvals}

    def __iter__(self):
        """
        Iterate through each of the ssoIds.
        """
        self.idx = 0
        return self

    def __next__(self):
        """
        Returns result of self._getObs when iterating over moSlicer.
        """
        if self.idx >= self.nSso:
            raise StopIteration
        idx = self.idx
        self.idx += 1
        return self._slice_obs(idx)

    def __getitem__(self, idx):
        # This may not be guaranteed to work if/when we implement chunking of the obs_file.
        return self._slice_obs(idx)

    def __eq__(self, other_slicer):
        """
        Evaluate if two slicers are equal.
        """
        result = False
        if isinstance(other_slicer, MoObjSlicer):
            if other_slicer.orbit_file == self.orbit_file:
                if other_slicer.obs_file == self.obs_file:
                    if np.array_equal(other_slicer.slice_points["H"], self.slice_points["H"]):
                        result = True
        return result



import warnings

from rubin_scheduler.skybrightness_pre import *

warnings.simplefilter("default")
warnings.warn(
    "rubin_sim.skybrightness_pre is deprecated, switch to rubin_scheduler.skybrightness_pre",
    DeprecationWarning,
)



import unittest

import numpy as np

from rubin_sim.maf.slicers import OneDSlicer, UniSlicer


def make_data_values(size=100, min=0.0, max=1.0, random=-1):
    """Generate a simple array of numbers,
    evenly arranged between min/max, but (optional) random order.
    """
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
    datavalues += min
    if random > 0:
        rng = np.random.RandomState(random)
        randorder = rng.rand(size)
        randind = np.argsort(randorder)
        datavalues = datavalues[randind]
    datavalues = np.array(list(zip(datavalues)), dtype=[("testdata", "float")])
    return datavalues


class TestUniSlicerSetupAndSlice(unittest.TestCase):
    def setUp(self):
        self.testslicer = UniSlicer()

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicertype(self):
        """Test instantiation of slicer sets slicer type as expected."""
        self.assertEqual(self.testslicer.slicer_name, self.testslicer.__class__.__name__)
        self.assertEqual(self.testslicer.slicer_name, "UniSlicer")

    def test_slicer_nbins(self):
        self.assertEqual(self.testslicer.nslice, 1)

    def test_setup_slicer_indices(self):
        """Test slicer returns correct indices (all) after setup.
        Note this also tests slicing.
        """
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        dv = make_data_values(nvalues, dvmin, dvmax, random=672)
        self.testslicer.setup_slicer(dv)
        # test slicing
        self.assertEqual(len(self.testslicer.indices), len(dv["testdata"]))
        np.testing.assert_equal(dv[self.testslicer.indices], dv)


class TestUniSlicerIteration(unittest.TestCase):
    def setUp(self):
        self.testslicer = UniSlicer()

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_iteration(self):
        """Test iteration -- which is a one-step identity op for a
        unislicer.
        """
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        dv = make_data_values(nvalues, dvmin, dvmax, random=432)
        self.testslicer.setup_slicer(dv)
        for i, b in enumerate(self.testslicer):
            pass
        self.assertEqual(i, 0)

    def test_get_item(self):
        """Test that can return an individual indexed values of the slicer."""
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        dv = make_data_values(nvalues, dvmin, dvmax, random=1192)
        self.testslicer.setup_slicer(dv)
        self.assertEqual(self.testslicer[0]["slice_point"]["sid"], 0)


class TestUniSlicerEqual(unittest.TestCase):
    def setUp(self):
        self.testslicer = UniSlicer()
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        dv = make_data_values(nvalues, dvmin, dvmax, random=3482)
        self.testslicer.setup_slicer(dv)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_equivalence(self):
        """Test equals method."""
        # Note that two uni slicers will be considered equal if they are
        # both the same kind of slicer (unislicer).
        # They will not necessarily slice data equally though (the indices are
        #  not necessarily the same!).
        # These should be the same, even though data is not the same.
        testslicer2 = UniSlicer()
        dv2 = make_data_values(100, 0, 1, random=43298)
        testslicer2.setup_slicer(dv2)
        self.assertEqual(self.testslicer, testslicer2)
        # these will not be the same, as different slicer type.
        testslicer2 = OneDSlicer(slice_col_name="testdata", bins=np.arange(0, 10, 1))
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)


if __name__ == "__main__":
    unittest.main()



import os
import sys
from typing import List

import lsst_sphinx_bootstrap_theme
from documenteer.sphinxconfig.utils import form_ltd_edition_name

# Work around Sphinx bug related to large and highly-nested source files
sys.setrecursionlimit(2000)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "sphinx_click.ext",
    "sphinxcontrib.autoprogram",
    "sphinx-prompt",
    "numpydoc",
    # "sphinx_autodoc_typehints",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "documenteer.sphinxext",
]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "rubin_sim/index"

# General information about the project.
project = "rubin_sim"
copyright = "2015-2022 " "Association of Universities for Research in Astronomy, Inc. (AURA)"
author = "LSST Survey Strategy Team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
if os.getenv("TRAVIS_BRANCH", default="main") == "main":
    # Use the current release as the version tag if on master
    version = "Current"
    release = version
else:
    # Use branch name as the version tag
    version = form_ltd_edition_name(git_ref_name=os.getenv("TRAVIS_BRANCH", default="main"))
    release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "README.rst"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# The reST default role cross-links Python (used for this markup: `text`)
default_role = "py:obj"

# Intersphinx

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "requests": ("https://2.python-requests.org/en/master/", None),
    "developer": ("https://developer.lsst.io/", None),
    "pybtex": ("https://docs.pybtex.org/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}

# Warnings to ignore
nitpick_ignore = [
    # This link to the base pybtex still never resolves because it is not
    # in pybtex's intersphinx'd API reference.
    ("py:class", "pybtex.style.formatting.plain.Style"),
]

# -- Options for linkcheck builder ----------------------------------------

linkcheck_retries = 2

# Since Jira is currently down at this time
linkcheck_ignore = [r"^https://jira.lsstcorp.org/browse/"]

linkcheck_timeout = 15

# -- Options for HTML output ----------------------------------------------

templates_path = [
    "_templates",
    lsst_sphinx_bootstrap_theme.get_html_templates_path(),
]

html_theme = "lsst_sphinx_bootstrap_theme"
html_theme_path = [lsst_sphinx_bootstrap_theme.get_html_theme_path()]


html_context = {
    # Enable "Edit in GitHub" link
    "display_github": True,
    # https://{{ github_host|default("github.com") }}/{{ github_user }}/
    #     {{ github_repo }}/blob/
    #     {{ github_version }}{{ conf_py_path }}{{ pagename }}{{ suffix }}
    "github_user": "lsst",
    "github_repo": "rubin_sim",
    "conf_py_path": "docs/",
    # TRAVIS_BRANCH is available in CI, but master is a safe default
    "github_version": os.getenv("TRAVIS_BRANCH", default="main") + "/",
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {"logotext": project}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "rubin_sim"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path: List[str] = []

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# -- Options for the API reference ----------------------------------------
numpydoc_show_class_members = False

# -- ReStructuredText epilog for common links/substitutions ---------------
rst_epilog = """
.. _conda-forge: https://conda-forge.org
.. _conda: https://conda.io/en/latest/index.html
"""



# Collection of utilities for MAF that relate to Opsim specifically.
__all__ = (
    "get_sim_data",
    "scale_benchmarks",
    "calc_coadded_depth",
)

import os
import sqlite3

import numpy as np
import pandas as pd
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine import make_url


def get_sim_data(
    db_con,
    sqlconstraint,
    dbcols,
    stackers=None,
    table_name=None,
    full_sql_query=None,
):
    """Query an opsim database for the needed data columns and run any required stackers.

    Parameters
    ----------
    db_con : `str` or SQLAlchemy connectable, or sqlite3 connection
        Filename to a sqlite3 file, or a connection object that can be used by pandas.read_sql
    sqlconstraint : `str` or None
        SQL constraint to apply to query for observations. Ignored if full_sql_query is set.
    dbcols : `list` [`str`]
        Columns required from the database. Ignored if full_sql_query is set.
    stackers : `list` [`rubin_sim.maf.stackers`], optional
        Stackers to be used to generate additional columns. Default None.
    table_name : `str` (None)
        Name of the table to query. Default None will try "observations" and "SummaryAllProps".
        Ignored if full_sql_query is set.
    full_sql_query : `str`
        The full SQL query to use. Overrides sqlconstraint, dbcols, tablename.

    Returns
    -------
    sim_data: `np.ndarray`
        A numpy structured array with columns resulting from dbcols + stackers, for observations matching
        the SQLconstraint.
    """
    if sqlconstraint is None:
        sqlconstraint = ""

    # Check that file exists
    if type(db_con) == str:
        if os.path.isfile(db_con) is False:
            raise FileNotFoundError("No file %s" % db_con)

    # Check if table is "observations" or "SummaryAllProps"
    if (table_name is None) & (full_sql_query is None) & (type(db_con) == str):
        url = make_url("sqlite:///" + db_con)
        eng = create_engine(url)
        inspector = inspect(eng)
        tables = [inspector.get_table_names(schema=schema) for schema in inspector.get_schema_names()]
        if "observations" in tables[0]:
            table_name = "observations"
        elif "SummaryAllProps" in tables[0]:
            table_name = "SummaryAllProps"
        elif "summary" in tables[0]:
            table_name = "summary"
        else:
            ValueError("Could not guess table_name, set with table_name or full_sql_query kwargs")
    elif (table_name is None) & (full_sql_query is None):
        # If someone passes in a connection object with an old table_name things will fail
        # that's probably fine, keep people from getting fancy with old sims
        table_name = "observations"

    if type(db_con) == str:
        con = sqlite3.connect(db_con)
    else:
        con = db_con

    if full_sql_query is None:
        col_str = ""
        for colname in dbcols:
            col_str += colname + ", "
        col_str = col_str[0:-2] + " "

        # Need to guess "observations" and "SummaryAllProps" for the table name
        # to be backwards compatible I guess
        query = "SELECT %s FROM %s" % (col_str, table_name)
        if len(sqlconstraint) > 0:
            query += " WHERE %s" % (sqlconstraint)
        query += ";"
        sim_data = pd.read_sql(query, con).to_records(index=False)

    else:
        query = full_sql_query
        sim_data = pd.read_sql(query, con).to_records(index=False)

    if len(sim_data) == 0:
        raise UserWarning("No data found matching sqlconstraint %s" % (sqlconstraint))
    # Now add the stacker columns.
    if stackers is not None:
        for s in stackers:
            sim_data = s.run(sim_data)
    return sim_data


def scale_benchmarks(run_length, benchmark="design"):
    """
    Set the design and stretch values of the number of visits, area of the footprint,
    seeing values, FWHMeff values, skybrightness, and single visit depth (based on SRD values).
    Scales number of visits for the length of the run, relative to 10 years.

    Parameters
    ----------
    run_length : float
        The length (in years) of the run.
    benchmark : str
        design or stretch - which version of the SRD values to return.
        requested is another option, in which case the values of the number of visits requested
        by the OpSim run (recorded in the Config table) is returned.

    Returns
    -------
    benchmarks: `dict` of floats
       A dictionary containing the number of visits, area of footprint, seeing and FWHMeff values,
       skybrightness and single visit depth for either the design or stretch SRD values.
    """
    # Set baseline (default) numbers for the baseline survey length (10 years).
    baseline = 10.0

    design = {}
    stretch = {}

    design["nvisitsTotal"] = 825
    stretch["nvisitsTotal"] = 1000
    design["Area"] = 18000
    stretch["Area"] = 20000

    design["nvisits"] = {"u": 56, "g": 80, "r": 184, "i": 184, "z": 160, "y": 160}
    stretch["nvisits"] = {"u": 70, "g": 100, "r": 230, "i": 230, "z": 200, "y": 200}

    # mag/sq arcsec
    design["skybrightness"] = {
        "u": 21.8,
        "g": 22.0,
        "r": 21.3,
        "i": 20.0,
        "z": 19.1,
        "y": 17.5,
    }
    stretch["skybrightness"] = {
        "u": 21.8,
        "g": 22.0,
        "r": 21.3,
        "i": 20.0,
        "z": 19.1,
        "y": 17.5,
    }

    # arcsec - old seeing values
    design["seeing"] = {"u": 0.77, "g": 0.73, "r": 0.7, "i": 0.67, "z": 0.65, "y": 0.63}
    stretch["seeing"] = {
        "u": 0.77,
        "g": 0.73,
        "r": 0.7,
        "i": 0.67,
        "z": 0.65,
        "y": 0.63,
    }

    # arcsec - new FWHMeff values (scaled from old seeing)
    design["FWHMeff"] = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }
    stretch["FWHMeff"] = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }

    design["singleVisitDepth"] = {
        "u": 23.9,
        "g": 25.0,
        "r": 24.7,
        "i": 24.0,
        "z": 23.3,
        "y": 22.1,
    }
    stretch["singleVisitDepth"] = {
        "u": 24.0,
        "g": 25.1,
        "r": 24.8,
        "i": 24.1,
        "z": 23.4,
        "y": 22.2,
    }

    # Scale the number of visits.
    if run_length != baseline:
        scalefactor = float(run_length) / float(baseline)
        # Calculate scaled value for design and stretch values of nvisits, per filter.
        for f in design["nvisits"]:
            design["nvisits"][f] = int(np.floor(design["nvisits"][f] * scalefactor))
            stretch["nvisits"][f] = int(np.floor(stretch["nvisits"][f] * scalefactor))

    if benchmark == "design":
        return design
    elif benchmark == "stretch":
        return stretch
    else:
        raise ValueError("Benchmark value %s not understood: use 'design' or 'stretch'" % (benchmark))


def calc_coadded_depth(nvisits, single_visit_depth):
    """
    Calculate the coadded depth expected for a given number of visits and single visit depth.

    Parameters
    ----------
    nvisits : dict of ints or floats
        Dictionary (per filter) of number of visits
    single_visit_depth : dict of floats
        Dictionary (per filter) of the single visit depth

    Returns
    -------
    dict of floats
        Dictionary of coadded depths per filter.
    """
    coadded_depth = {}
    for f in nvisits:
        if f not in single_visit_depth:
            raise ValueError("Filter keys in nvisits and single_visit_depth must match")
        coadded_depth[f] = float(1.25 * np.log10(nvisits[f] * 10 ** (0.8 * single_visit_depth[f])))
        if not np.isfinite(coadded_depth[f]):
            coadded_depth[f] = single_visit_depth[f]
    return coadded_depth



# pylint: disable=too-many-arguments
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name

# imports
import unittest

import numpy as np
import pandas as pd

from rubin_sim.maf.slicers import BlockIntervalSlicer, TimeIntervalSlicer, VisitIntervalSlicer

# constants

SIMDATA = {
    "observationStartMJD": [
        60000.100,
        60000.101,
        60000.102,
        60000.103,
        60000.202,
        60000.203,
    ],
    "visitTime": [86, 86, 86, 86, 86, 86],
    "note": [
        "greedy",
        "blob, yy, a",
        "blob, yy, a",
        "DD:EDFS, b",
        "DD:EDFS, b",
        "DD:EDFS, b",
    ],
}

# interface functions

# classes


class TestTimeIntervalSlicer(unittest.TestCase):
    interval_seconds = 60

    def setUp(self):
        self.slicer = TimeIntervalSlicer(interval_seconds=self.interval_seconds)

    def test_setup_slicer(self):
        self.slicer.setup_slicer(SIMDATA)
        self.assertEqual(self.slicer.nslice, 24 * 60)
        slice_points = self.slicer.get_slice_points()
        nonempty_slices = [item[0] for item in self.slicer.sim_idxs.items() if len(item[1]) > 0]
        self.assertTrue(
            np.allclose(
                np.array(SIMDATA["observationStartMJD"]),
                slice_points["mjd"][nonempty_slices],
                atol=self.interval_seconds / (24 * 60 * 60.0),
            )
        )
        self.assertTrue(np.all(slice_points["duration"] == self.interval_seconds))


class TestBlockIntervalSlicer(unittest.TestCase):
    def setUp(self):
        self.slicer = BlockIntervalSlicer()

    def test_setup_slicer(self):
        self.slicer.setup_slicer(SIMDATA)
        self.assertEqual(self.slicer.nslice, 4)
        slice_points = self.slicer.get_slice_points()
        sim_idxs = self.slicer.sim_idxs
        visits = pd.DataFrame(SIMDATA)
        for sid in slice_points["sid"]:
            these_visits = visits.iloc[sim_idxs[sid]]
            self.assertTrue(np.all(these_visits["note"] == these_visits["note"].values[0]))


class TestVisitIntervalSlicer(unittest.TestCase):
    def setUp(self):
        self.slicer = VisitIntervalSlicer()

    def test_setup_slicer(self):
        self.slicer.setup_slicer(SIMDATA)
        self.assertEqual(self.slicer.nslice, len(SIMDATA["observationStartMJD"]))
        slice_points = self.slicer.get_slice_points()
        self.assertIn("sid", slice_points)
        self.assertIn("mjd", slice_points)
        self.assertIn("duration", slice_points)
        self.assertTrue(np.all(slice_points["duration"] == SIMDATA["visitTime"]))


# internal functions & classes

run_tests_now = __name__ == "__main__"
if run_tests_now:
    unittest.main()



__all__ = ("ddf_dir",)

import matplotlib

matplotlib.use("Agg")
import argparse
import glob
import os
import shutil

import rubin_sim.maf.batches as batches
import rubin_sim.maf.db as db
import rubin_sim.maf.metric_bundles as mb


def ddf_dir():
    """
    Run the glance batch on all .db files in a directory.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--db", type=str, default=None)
    parser.add_argument("--nside", type=int, default=512)
    args = parser.parse_args()

    if args.db is None:
        db_files = glob.glob("*.db")
        db_files = [filename for filename in db_files if "trackingDb" not in filename]
    else:
        db_files = [args.db]
    run_names = [os.path.basename(name).replace(".db", "") for name in db_files]

    for filename, name in zip(db_files, run_names):
        if os.path.isdir(name + "_ddf"):
            shutil.rmtree(name + "_ddf")

        bdict = {}
        bdict.update(batches.ddfBatch(run_name=name, nside=args.nside))
        results_db = db.ResultsDb(out_dir=name + "_ddf")
        group = mb.MetricBundleGroup(
            bdict,
            filename,
            out_dir=name + "_ddf",
            results_db=results_db,
            save_early=False,
        )
        group.run_all(clear_memory=True, plot_now=True)
        results_db.close()
        db.add_run_to_database(name + "_ddf", "trackingDb_sqlite.db", None, name, "", "", name + ".db")
